import abc
import sys
import os
import warnings
from collections import OrderedDict
from functools import partial
from typing import Optional

import numpy as np
import six
import yaml

from .container import DataContainerBase
from ..io.file import FileIOMixin
from ...config import kc
from ...core.fitters.nexus import Nexus, Parameter, Array
from ...core.fitters.nexus_fitter import NexusFitter
from ...core.constraint import GaussianMatrixParameterConstraint, GaussianSimpleParameterConstraint
from ...core.error import CovMat
from ...tools import print_dict_as_table
from .._base.cost import CostFunction, STRING_TO_COST_FUNCTION
from ..util import invert_matrix, is_diagonal, cholesky_decomposition, log_determinant, \
    log_determinant_pointwise, to_python_types, to_numpy_arrays

__all__ = ["FitBase"]


@six.add_metaclass(abc.ABCMeta)
class FitBase(FileIOMixin, object):
    """
    This is a purely abstract class implementing the minimal interface required by all
    types of fitters.
    """

    CONTAINER_TYPE = DataContainerBase
    MODEL_TYPE = None
    MODEL_FUNCTION_TYPE = None
    PLOT_ADAPTER_TYPE = None
    RESERVED_NODE_NAMES = set()
    _BASIC_ERROR_NAMES = {}
    _STRING_TO_COST_FUNCTION = STRING_TO_COST_FUNCTION
    _AXES = (None,)  # axes for which to for example create data nexus nodes
    _MODEL_NAME = "model"
    _MODEL_ERROR_NODE_NAMES = ["model_error", "model_cov_mat"]

    def __init__(
            self, data, model_function, cost_function, minimizer=None, minimizer_kwargs=None,
            dynamic_error_algorithm="nonlinear"):
        """This is a purely abstract class implementing the minimal interface required by all
        types of fits.

        :param minimizer: Name of the minimizer to use.
        :type minimizer: str or None
        :param minimizer_kwargs: Dictionary wit keywords for initializing the minimizer.
        :type minimizer_kwargs: dict or None
        :param dynamic_error_algorithm: how to handle errors that depend on the model parameters.
        :type dynamic_error_algorithm: "nonlinear" or "iterative".
        """
        super(FitBase, self).__init__()
        self._data_container = None
        self._param_model = None
        self._nexus = None
        self._fitter = None
        self._fit_param_names = []  # names of all fit parameters
        self._fit_param_constraints = []
        self._loaded_result_dict = None  # contains potential fit results from a file or multifit

        # save minimizer, minimizer_kwargs for serialization
        self._minimizer = minimizer
        self._minimizer_kwargs = minimizer_kwargs

        self.dynamic_error_algorithm = dynamic_error_algorithm
        self._dynamic_error_warning_printed = False

        # set/construct the model function object
        if self.MODEL_FUNCTION_TYPE is None:
            if model_function is not None:
                raise ValueError(
                    "Model function must be None if subclass does not define MODEL_FUNCTION_TYPE.")
            self._model_function = model_function
        else:
            if isinstance(model_function, self.MODEL_FUNCTION_TYPE):
                self._model_function = model_function
            else:
                self._model_function = self.MODEL_FUNCTION_TYPE(model_function)

            # Disallow using reserved keywords as model function arguments:
            if not self.RESERVED_NODE_NAMES.isdisjoint(
                    set(self._model_function.signature.parameters)):
                _invalid_args = self.RESERVED_NODE_NAMES.intersection(
                    set(self._model_function.signature.parameters))
                raise ValueError(
                    "The following names are reserved and cannot be used as model function arguments: %r"
                    % (_invalid_args,))

        # set and validate the cost function
        self._implicit_no_errors = cost_function == "chi2" and isinstance(data, DataContainerBase) \
            and not data.has_errors
        if self._implicit_no_errors:
            cost_function = "chi2_no_errors"
        if isinstance(cost_function, CostFunction):
            self._cost_function = cost_function
        elif isinstance(cost_function, str):
            try:
                _cost_function_class, _kwargs = self._STRING_TO_COST_FUNCTION[cost_function]
            except KeyError:
                raise ValueError('Unknown cost function: %s' % cost_function)
            self._cost_function = _cost_function_class(**_kwargs)
        elif cost_function is not None:
            self._cost_function = CostFunction(cost_function)
            # self._validate_cost_function_raise()
            # TODO: validate user-defined cost function? how?
        else:
            self._cost_function = None
        self._cost_function_pointwise = None if self._cost_function is None \
            else self._cost_function.pointwise_version

        # initialize the Nexus
        self._init_nexus()

        # initialize the Fitter
        self._initialize_fitter()

        # set the data after the cost_function has been set and nexus has been initialized
        if data is not None:
            self.data = data

    # -- private methods

    def _add_property_to_nexus(self, prop, obj=None, name=None, depends_on=None):
        """register a property of this object in the nexus as a function node"""
        obj = obj if obj is not None else self
        _node = self._nexus.add_function(
            partial(getattr(obj.__class__, prop).fget, obj),
            func_name=name or prop,
            par_names=[]
        )
        if depends_on is not None:
            self._nexus.add_dependency(name=_node.name, depends_on=depends_on)
        return _node

    @classmethod
    def _get_base_class(cls):
        return FitBase

    @classmethod
    def _get_object_type_name(cls):
        return 'fit'

    def _init_nexus(self):
        self._nexus = Nexus()

        # -- fit parameters

        _parameter_nodes = []
        # get names and default values of all parameters if the fit has a model function
        if self._model_function is not None:
            for _par_name, _par_value in six.iteritems(self._model_function.defaults_dict):
                # create nexus node for function parameter
                _parameter_nodes.append(self._nexus.add(Parameter(_par_value, name=_par_name)))

                self._fit_param_names.append(_par_name)

        self._nexus.add(Array(_parameter_nodes, name="parameter_values"))
        self._nexus.add_function(lambda: self.parameter_constraints,
                                 func_name='parameter_constraints', par_names=[])

        # -- errors

        for _axis in self._AXES:
            _error_names = []
            _mat_names = []
            for _type in ("model", "data", "total"):
                _name = '_'.join((_axis, _type)) if _axis is not None else _type
                _error_name = '_'.join((_axis, _type, 'error')) if _axis is not None \
                    else '_'.join((_type, 'error'))
                _mat_name = '_'.join((_axis, _type, "cov_mat")) if _axis is not None \
                    else '_'.join((_type, "cov_mat"))
                if _type == "total":
                    # Performance optimization: calculate total foo in Nexus
                    self._nexus.add_function(
                        lambda m, d: np.sqrt(m ** 2 + d ** 2),
                        func_name=_error_name, par_names=_error_names)
                    self._nexus.add_function(
                        lambda m, d: m + d, func_name=_mat_name, par_names=_mat_names)
                else:
                    self._add_property_to_nexus(_name)
                    self._add_property_to_nexus(_error_name)
                    _error_names.append(_error_name)
                    self._add_property_to_nexus(_mat_name)
                    _mat_names.append(_mat_name)
                if _error_name in self._MODEL_ERROR_NODE_NAMES:
                    self._nexus.add_dependency(_error_name, depends_on="parameter_values")
                if _mat_name in self._MODEL_ERROR_NODE_NAMES:
                    self._nexus.add_dependency(_mat_name, depends_on="parameter_values")
                self._add_property_to_nexus(_mat_name + "_inverse", depends_on=_mat_name)
                self._nexus.add_function(cholesky_decomposition, func_name=_mat_name + "_cholesky",
                                         par_names=[_mat_name])
                self._nexus.add_function(log_determinant, func_name=_mat_name + "_log_determinant",
                                         par_names=[_mat_name + "_cholesky"])
                self._nexus.add_function(
                    log_determinant_pointwise,
                    func_name=_error_name + "_squared_log_sum",
                    par_names=[_error_name]
                )

        if self._model_function is not None:
            # add the original function name as an alias to 'model'
            try:
                self._nexus.add_alias(self._model_function.name, alias_for=self._MODEL_NAME)
            except ValueError:
                pass  # allow 'model' as function name for model
        self._init_cost_function()


    def _init_cost_function(self, existing_behavior='fail'):
        if self._cost_function is not None:
            # the cost function (the function to be minimized)
            _cost_node = self._nexus.add_function(
                self._cost_function,
                par_names=self._cost_function.arg_names,
                func_name=self._cost_function.name,
                existing_behavior=existing_behavior
            )

            if self._cost_function.name != "cost":
                _cost_alias = self._nexus.add_alias(
                    'cost', alias_for=self._cost_function.name, existing_behavior=existing_behavior)
        if self._cost_function_pointwise is not None:
            self._nexus.add_function(
                self._cost_function_pointwise,
                par_names=self._cost_function_pointwise.arg_names,
                func_name=self._cost_function_pointwise.name,
                existing_behavior=existing_behavior
            )

    def _initialize_fitter(self):
        self._fitter = NexusFitter(nexus=self._nexus,
                                   parameters_to_fit=self._fit_param_names,
                                   parameter_to_minimize=self._cost_function.name,
                                   minimizer=self._minimizer,
                                   minimizer_kwargs=self._minimizer_kwargs)

    @abc.abstractmethod
    def _set_new_data(self, new_data):
        """Private method called by the :py:attr:`~data` setter. Must be overwritten for each derived class."""
        pass

    @abc.abstractmethod
    def _set_new_parametric_model(self):
        pass

    # Overridden by MultiFit
    def _get_model_function_argument_formatters(self):
        """All arguments of the model function including independent variables."""
        return self._model_function.formatter.arg_formatters

    # Overridden by MultiFit
    def _get_model_function_parameter_formatters(self):
        """Only the parameters which are fitted. Excludes independent variables."""
        return self.model_function.formatter.par_formatters

    def _report_data(self, output_stream, indent, indentation_level):
        """Report the data used in this fit to the given output stream.

        :param io.TextIOBase output_stream: The output stream to which the report should be printed.
        :param str indent: The string to use when indenting lines.
        :param int indentation_level: How many times the indent str should be placed before each line.
        """
        pass

    def _report_model(self, output_stream, indent, indentation_level):
        """Report the fit model to the given output stream.

        :param io.TextIOBase output_stream: The output stream to which the report should be printed.
        :param str indent: The string to use when indenting lines.
        :param int indentation_level: How many times the indent str should be placed before each line.
        """
        output_stream.write(indent * indentation_level + '#########\n')
        output_stream.write(indent * indentation_level + '# Model #\n')
        output_stream.write(indent * indentation_level + '#########\n\n')
        output_stream.write(indent * (indentation_level + 1) + "Model Function\n")
        output_stream.write(indent * (indentation_level + 1) + "==============\n\n")
        output_stream.write(indent * (indentation_level + 2))
        output_stream.write(self._model_function.formatter.get_formatted(with_expression=True))
        output_stream.write('\n\n')

    def _report_fit_results(self, output_stream, indent, indentation_level, asymmetric_parameter_errors):
        """Report the fit results

        :param io.TextIOBase output_stream: The output stream to which the report should be printed.
        :param str indent: The string to use when indenting lines.
        :param int indentation_level: How many times the indent str should be placed before each line.
        :param bool asymmetric_parameter_errors: If asymmetric parameter uncertainties should be used instead of
            symmetric uncertainties.
        """
        output_stream.write(indent * indentation_level + '###############\n')
        output_stream.write(indent * indentation_level + '# Fit Results #\n')
        output_stream.write(indent * indentation_level + '###############\n\n')

        if not self.did_fit:
            output_stream.write(indent * (indentation_level + 1) +
                                'WARNING: No fit has been performed as of yet. Did you forget to run fit.do_fit()?\n\n')

        output_stream.write(indent * (indentation_level + 1) + "Model Parameters\n")
        output_stream.write(indent * (indentation_level + 1) + "================\n\n")

        self._update_parameter_formatters(update_asymmetric_errors=asymmetric_parameter_errors)
        for _pf in self._get_model_function_parameter_formatters():
            output_stream.write(indent * (indentation_level + 2))
            output_stream.write(
                _pf.get_formatted(with_name=True,
                                  with_value=True,
                                  with_errors=self.errors_valid,
                                  format_as_latex=False,
                                  asymmetric_error=asymmetric_parameter_errors)
            )
            output_stream.write('\n')
        output_stream.write('\n')

        if self.errors_valid:
            output_stream.write(indent * (indentation_level + 1) + "Model Parameter Correlations\n")
            output_stream.write(indent * (indentation_level + 1)
                                + "============================\n\n")

            _cor_mat_content = self.parameter_cor_mat
            if _cor_mat_content is not None:
                par_display_names = [_pf.name for _pf in
                                    self._get_model_function_parameter_formatters()]
                _cor_mat_as_dict = OrderedDict()
                _cor_mat_as_dict['_invisible_first_column'] = par_display_names
                for _par_name, _row in zip(par_display_names, self.parameter_cor_mat.T):
                    _cor_mat_as_dict[_par_name] = np.atleast_1d(np.squeeze(np.asarray(_row)))

                print_dict_as_table(_cor_mat_as_dict, output_stream=output_stream, indent_level=2)
            else:
                output_stream.write(indent * (indentation_level + 2) + '<not available>\n')
            output_stream.write('\n')

        output_stream.write(indent * (indentation_level + 1) + "Cost Function\n")
        output_stream.write(indent * (indentation_level + 1) + "=============\n\n")

        _pf = self._cost_function.formatter
        output_stream.write(
            indent * (indentation_level + 2) + "Cost function: {}\n\n".format(_pf.description))
        _gof_value = self.goodness_of_fit
        output_stream.write(indent * (indentation_level + 2))
        if _gof_value is None or not self.errors_valid:
            output_stream.write("Cost = ")
            output_stream.write(_pf.get_formatted(
                value=self.cost_function_value,
                with_name=False,
                format_as_latex=False
            ))
        else:
            _gof_string = "chi2 / ndf = " if self._cost_function.is_chi2 else "GoF / ndf = "
            output_stream.write(_gof_string)
            output_stream.write(_pf.get_formatted(
                value=_gof_value,
                n_degrees_of_freedom=self.ndf,
                with_name=False,
                with_value_per_ndf=True,
                format_as_latex=False
            ))
        output_stream.write('\n\n')
        _chi2_prob = self.chi2_probability
        if _chi2_prob is not None and self.errors_valid:
            output_stream.write("%schi2 probability = %#.3g\n\n" % (
                indent * (indentation_level + 2), _chi2_prob))

    def _update_parameter_formatters(self, update_asymmetric_errors=False):
        """Update all parameter formatters with the current values and uncertainties.

        :param bool update_asymmetric_errors: If the asymmetric parameter uncertainties should be updated as well.
        """
        for _fpf, _pv, _pe in zip(
                self._get_model_function_parameter_formatters(), self.parameter_values, self.parameter_errors):
            _fpf.value = _pv
            _fpf.error = _pe
        if update_asymmetric_errors:
            self._check_dynamic_error_compatibility()
            _asymmetric_parameter_errors = self.asymmetric_parameter_errors
            if _asymmetric_parameter_errors is None:
                _asymmetric_parameter_errors = np.stack(
                    [-self.parameter_errors, self.parameter_errors], axis=1)
            for _fpf, _ape in zip(self._get_model_function_parameter_formatters(),
                                  _asymmetric_parameter_errors):
                _fpf.asymmetric_error = _ape

    def _on_error_change(self):
        """Mark all error nodes in :py:attr:`~_BASIC_ERROR_NAMES` for updates in the nexus."""
        self._fitter.reset_minimizer()
        for _error_name in self._BASIC_ERROR_NAMES:
            self._nexus.get(_error_name).mark_for_update()
        if self._implicit_no_errors:
            _cost_function_class, _kwargs = self._STRING_TO_COST_FUNCTION["chi2_covariance"]
            self._cost_function = _cost_function_class(**_kwargs)
            self._cost_function_pointwise = self._cost_function.pointwise_version
            self._init_cost_function(existing_behavior='replace')
            self._implicit_no_errors = False


    def _set_data_as_model_ref(self):
        for _err in self._param_model.get_matching_errors({"relative": True}).values():
            _old_ref = _err.reference
            _err.reference = self._data_container.data

    def _iterative_fits_needed(self):
        return bool(self._param_model.get_matching_errors({"relative": True})) \
               and self._dynamic_error_algorithm == "iterative"

    def _second_fit_needed(self):
        return bool(self._param_model.get_matching_errors({"relative": True})) \
               and self._dynamic_error_algorithm == "nonlinear"

    def _get_node_names_to_freeze(self, first_fit):
        if first_fit or not self._param_model.get_matching_errors({"relative": True}) \
                or self._dynamic_error_algorithm == "iterative":
            return self._MODEL_ERROR_NODE_NAMES
        else:
            return []

    def _pre_fit_iteration(self, first_fit=False):
        for _model_err_name in self._get_node_names_to_freeze(first_fit):
            _node = self._nexus.get(_model_err_name)
            _node.update()
            _node.freeze()

    def _post_fit_iteration(self, first_fit=False):
        for _model_err_name in self._get_node_names_to_freeze(first_fit):
            _node = self._nexus.get(_model_err_name)
            _node.unfreeze()
            _node.update()
            _node.notify_parents()

    def _check_dynamic_error_compatibility(self):
        if not self._dynamic_error_warning_printed and self._iterative_fits_needed():
            warnings.warn(
                "Asymmetric parameter errors, parameter profiles, and contours cannot be "
                "calculated with iterative dynamic errors. Used nonlinear errors instead.")
            self._dynamic_error_warning_printed = True


    def _get_default_x_ticks(self):
        return []

    # -- public properties

    @property
    def data(self):
        """array of measurement values"""
        return self._data_container.data

    @data.setter
    def data(self, new_data):
        self._set_new_data(new_data)
        # validate cost function
        _data_and_cost_compatible, _reason = self._cost_function.is_data_compatible(self.data)
        if not _data_and_cost_compatible:
            raise ValueError('Fit data and cost function are not compatible: %s' % _reason)
        self._set_new_parametric_model()
        self._param_model._on_error_change_callbacks = [self._on_error_change]

    @property
    def data_error(self):
        """array of pointwise data uncertainties"""
        return self._data_container.err

    @property
    def data_cov_mat(self):
        """the data covariance matrix"""
        return self._data_container.cov_mat

    @property
    def data_cov_mat_inverse(self):
        """inverse of the data covariance matrix (or ``None`` if singular)"""
        return self._data_container.cov_mat_inverse

    @property
    def data_cor_mat(self):
        """the data correlation matrix"""
        return self._data_container.cor_mat

    @property
    def data_container(self):
        """The data container used in this fit.

        :rtype: kafe2.fit._base.DataContainerBase
        """
        return self._data_container

    @property
    @abc.abstractmethod
    def model(self):
        pass

    @property
    def model_error(self):
        """array of pointwise model uncertainties"""
        self._param_model.parameters = self.parameter_values  # this is lazy, so just do it
        return self._param_model.err

    @property
    def model_cov_mat(self):
        """the model covariance matrix"""
        self._param_model.parameters = self.parameter_values  # this is lazy, so just do it
        return self._param_model.cov_mat

    @property
    def model_cov_mat_inverse(self):
        """inverse of the model covariance matrix (or ``None`` if singular)"""
        self._param_model.parameters = self.parameter_values  # this is lazy, so just do it
        return self._param_model.cov_mat_inverse

    @property
    def model_cor_mat(self):
        """the model correlation matrix"""
        self._param_model.parameters = self.parameter_values  # this is lazy, so just do it
        return self._param_model.cor_mat

    @property
    def total_error(self):
        """array of pointwise total uncertainties"""
        return self._nexus.get("total_error").value

    @property
    def total_cov_mat(self):
        """the total covariance matrix"""
        return self._nexus.get("total_cov_mat").value

    @property
    def total_cov_mat_inverse(self):
        """inverse of the total covariance matrix (or ``None`` if singular)"""
        return invert_matrix(self.total_cov_mat)

    @property
    def total_cor_mat(self):
        """the total correlation matrix"""
        _cov_mat = self.total_cov_mat
        if _cov_mat is None:
            return None
        else:
            return CovMat(self.total_cov_mat).cor_mat

    @property
    def model_function(self):
        """The wrapped model function as a :py:obj:`~kafe2.fit._base.ModelFunctionBase` or derived object.
        This object contains the model function as well as formatting information used for this fit.

        :rtype: kafe2.fit._base.ModelFunctionBase"""
        return self._model_function

    @property
    def model_label(self):
        """The label of the model used in this fit.

        :rtype: str or None
        """
        return self._param_model.label

    @model_label.setter
    def model_label(self, label):
        self._param_model.label = label

    @property
    def parameter_values(self):
        """The current parameter values.

        :rtype: numpy.ndarray[float]
        """
        return self._nexus.get("parameter_values").value

    @property
    def parameter_names(self):
        """The current parameter names.

        :rtype: tuple[str]
        """
        return self._fitter.parameters_to_fit

    @property
    def parameter_errors(self):
        """
        The current parameter uncertainties.
        Can be set to control initial step size during minimization.

        :rtype: numpy.ndarray[float]
        """
        if self._loaded_result_dict is not None:
            return self._loaded_result_dict['parameter_errors']
        return self._fitter.fit_parameter_errors

    @parameter_errors.setter
    def parameter_errors(self, new_errors):
        self._fitter.fit_parameter_errors = new_errors

    @property
    def parameter_cov_mat(self):
        """The current parameter covariance matrix.

        :rtype: None or numpy.ndarray[numpy.ndarray[float]]
        """
        if self._loaded_result_dict is not None:
            return self._loaded_result_dict['parameter_cov_mat']
        return self._fitter.fit_parameter_cov_mat

    @property
    def parameter_cor_mat(self):
        """The current parameter correlation matrix.

        :rtype: None or numpy.ndarray[numpy.ndarray[float]]
        """
        if self._loaded_result_dict is not None:
            return self._loaded_result_dict['parameter_cor_mat']
        return self._fitter.fit_parameter_cor_mat

    @property
    def asymmetric_parameter_errors(self):
        """The current asymmetric parameter uncertainties.

        :rtype: numpy.ndarray[numpy.ndarray[float, float]]
        """
        if self._loaded_result_dict is not None and self._loaded_result_dict['asymmetric_parameter_errors'] is not None:
            return self._loaded_result_dict['asymmetric_parameter_errors']
        self._check_dynamic_error_compatibility()
        return self._fitter.asymmetric_fit_parameter_errors

    @property
    def parameter_name_value_dict(self):
        """A dictionary mapping each parameter name to its current value.

        :rtype: OrderedDict[str, float]
        """
        return self._fitter.get_fit_parameter_values()

    @property
    def parameter_constraints(self):
        """The gaussian constraints given for the fit parameters.

        :rtype: list[kafe2.core.constraint.GaussianSimpleParameterConstraint or
            kafe2.core.constraint.GaussianMatrixParameterConstraint]
        """
        return self._fit_param_constraints

    @property
    def cost_function_value(self):
        """The current value of the cost function.

        :rtype: float
        """
        return self._fitter.parameter_to_minimize_value

    @property
    def data_size(self):
        """The size (number of points) of the data container.

        :rtype: int
        """
        return self._data_container.size

    @property
    def has_model_errors(self):
        """:py:obj:`True` if at least one uncertainty source is defined for the model.

        :rtype: bool
        """
        return self._param_model is not None and self._param_model.has_errors

    @property
    def has_data_errors(self):
        """:py:obj:`True` if at least one uncertainty source is defined for the data.

        :rtype: bool
        """
        return self._data_container is not None and self._data_container.has_errors

    @property
    def has_errors(self):
        """:py:obj:`True` if at least one uncertainty source is defined for either the data or the model.

        :rtype: bool
        """
        return self.has_data_errors or self.has_model_errors

    @property
    def model_count(self):
        """The number of model functions contained in the fit, 1 by default.

        :rtype: int
        """
        return 1

    @property
    def did_fit(self):
        """Whether a fit was performed for the given data and model.

        :rtype: bool
        """
        if self._loaded_result_dict is not None:
            return self._loaded_result_dict['did_fit']
        return self._fitter.state_is_from_minimizer

    @property
    def ndf(self):
        """The degrees of freedom of this fit.

        :rtype: int
        """
        _extra_ndf_constraints = 0
        for _parameter_constraint in self._fit_param_constraints:
            _extra_ndf_constraints += _parameter_constraint.extra_ndf
        return self._param_model.ndf + len(self._fitter.fixed_parameters) + _extra_ndf_constraints

    @property
    def goodness_of_fit(self):
        if self._cost_function_pointwise is not None and is_diagonal(self.total_cov_mat):
            _cost_function = self._cost_function_pointwise
        else:
            _cost_function = self._cost_function
        return _cost_function.goodness_of_fit(
            *[self._nexus.get(_node_name).value for _node_name in _cost_function.arg_names])

    @property
    def dynamic_error_algorithm(self):
        """The algorithm to use for handling errors that depend on the model parameters.
        :rtype: str
        """
        return self._dynamic_error_algorithm

    @dynamic_error_algorithm.setter
    def dynamic_error_algorithm(self, new_dea):
        _valid_deas = ["nonlinear", "iterative"]
        if new_dea not in _valid_deas:
            raise ValueError(
                "Unknown dynamic error algorithm: %s. Valid algorithms: %s" % (
                    new_dea, _valid_deas))
        self._dynamic_error_algorithm = new_dea

    @property
    def chi2_probability(self):
        """The chi2 probability for the current model values."""
        _cost = self.cost_function_value
        if self._cost_function.add_determinant_cost:
            _cost -= self._nexus.get("total_cov_mat_log_determinant").value
        return self._cost_function.chi2_probability(_cost, self.ndf)

    @property
    def errors_valid(self):
        return self._cost_function.errors_valid and self.did_fit and (
            self.has_errors or not self._cost_function.needs_errors)
    # public methods

    def set_parameter_values(self, **param_name_value_dict):
        """Set the fit parameters to new values. Valid keyword arguments are the names of the declared fit parameters.

        :param param_name_value_dict: new parameter values
        """
        _return_value = self._fitter.set_fit_parameter_values(**param_name_value_dict)
        if self._param_model is not None:
            self._param_model.parameters = self.parameter_values
        return _return_value

    def set_all_parameter_values(self, param_value_list):
        """Set all the fit parameters at the same time.

        :param typing.Iterable[float] param_value_list: List of parameter values (mind the order).
        """
        if self._param_model is not None:
            self._param_model.parameters = param_value_list
        return self._fitter.set_all_fit_parameter_values(param_value_list)

    def fix_parameter(self, name, value=None):
        """Fix a parameter so that its value doesn't change when calling :py:meth:`~do_fit()`.

        :param str name: The name of the parameter to be fixed
        :param float or None value: The value to be given to the fixed parameter. If :py:obj:`None` the current value
            from :py:attr:`~parameter_values` will be used.
        """
        self._fitter.fix_parameter(name=name, value=value)
        _par_index = self.parameter_names.index(name)
        self._get_model_function_parameter_formatters()[_par_index].fixed = True

    def release_parameter(self, name):
        """Release a fixed parameter so that its value once again changes when calling :py:meth:`~do_fit()`.

        :param str name: The name of the fixed parameter to be released
        """
        self._fitter.release_parameter(name=name)
        _par_index = self.parameter_names.index(name)
        self._get_model_function_parameter_formatters()[_par_index].fixed = False

    def limit_parameter(self, name, lower=None, upper=None):
        """Limit a parameter to a given range.

        :param str name: The name of the parameter to limit.
        :param float lower: The minimum parameter value.
        :param float upper: The maximum parameter value.
        """
        if lower is None and upper is None:
            raise ValueError("Either a lower or an upper bound must be provided!")

        # make sure these are numeric values (otherwise minimizer will
        # fail on `do_fit` with a cryptic error)
        for _lim in (lower, upper):
            try:
                assert _lim is None or float(_lim) == _lim
            except (TypeError, ValueError, AssertionError):
                six.raise_from(
                    TypeError("Expecting `None` or numeric value for parameter limit, got {}: {}".format(type(_lim), repr(_lim))),
                    None
                )

        self._fitter.limit_parameter(name=name, limits=(lower, upper))

    def unlimit_parameter(self, name):
        """Unlimit a parameter.

        :param str name: The name of the parameter to unlimit.
        """
        self._fitter.unlimit_parameter(name=name)

    def add_matrix_parameter_constraint(self, names, values, matrix, matrix_type='cov', uncertainties=None,
                                        relative=False):
        """Advanced class for applying correlated constraints to several parameters of a fit.
        The order of **names**, **values**, **matrix**, and **uncertainties** must be aligned.
        In other words the first index must belong to the first value, the first row/column in the matrix, etc.

        Let N be the number of parameters to be constrained.

        :param names: The names of the parameters to be constrained. Must be of shape (N,).
        :type names: typing.Collection[str]
        :param values: The values to which the parameters should be constrained. Must be of shape shape (N,).
        :type values: typing.Sized[float]
        :param matrix: The matrix that defines the correlation between the parameters. By default interpreted as a
            covariance matrix. Can also be interpreted as a correlation matrix by setting **matrix_type**.
            Must be of shape shape (N, N).
        :type matrix: typing.Iterable[float]
        :param matrix_type: Either ``'cov'`` or ``'cor'``. Defines whether the matrix should be interpreted as a
            covariance matrix or as a correlation matrix.
        :type matrix_type: str
        :param uncertainties: The uncertainties to be used in conjunction with a correlation matrix.
            Must be of shape (N,)
        :type uncertainties: None or typing.Iterable[float]
        :param relative: Whether the covariance matrix/the uncertainties should be interpreted as relative to
            **values**.
        :type relative: bool
        """
        if len(names) != len(values):
            raise ValueError(
                'Lengths of names and values are different: %s <-> %s' % (len(names), len(values)))
        _par_indices = []
        for _name in names:
            try:
                _par_indices.append(self.parameter_names.index(_name))
            except ValueError as _e:
                raise ValueError('Unknown parameter name: %s' % _name) from _e
        self._fit_param_constraints.append(GaussianMatrixParameterConstraint(
            indices=_par_indices, values=values, matrix=matrix, matrix_type=matrix_type, uncertainties=uncertainties,
            relative=relative
        ))

    def add_parameter_constraint(self, name, value, uncertainty, relative=False):
        """Apply a simple gaussian constraint to a single fit parameter.

        :param str name: The name of the parameter to be constrained.
        :param float value: The value to which the parameter should be constrained.
        :param float uncertainty: The uncertainty with which the parameter should be constrained to the given value.
        :param bool relative: Whether the given uncertainty is relative to the given value.
        """
        try:
            _index = self.parameter_names.index(name)
        except ValueError as _e:
            raise ValueError('Unknown parameter name: %s' % name) from _e
        self._fit_param_constraints.append(GaussianSimpleParameterConstraint(
            index=_index, value=value, uncertainty=uncertainty, relative=relative
        ))

    def get_matching_errors(self, matching_criteria=None, matching_type='equal'):
        """Return a list of uncertainty objects fulfilling the specified matching criteria.

        Valid keys for **matching_criteria**:
            * ``name`` (the unique error name)
            * ``type`` (either ``'simple'`` or ``'matrix'``)
            * ``correlated`` (bool, only matches simple errors!)
            * ``reference`` (either ``'model'`` or ``'data'``)

        .. note::
            The error objects contained in the dictionary are not copies, but the original error objects.
            Modifying them is possible, but not recommended.
            If you do modify any of them, the changes will not be reflected in the total error calculation until the
            error cache is cleared. This can be done by calling the private dataset method
            :py:meth:`~kafe2.fit._base.DataContainerBase._clear_total_error_cache`.

        :param matching_criteria: Key-value pairs specifying matching criteria. The resulting error array will only
                                  contain error objects matching *all* provided criteria.
                                  If :py:obj:`None`, all error objects are returned.
        :type matching_criteria: dict or None
        :param matching_type: How to perform the matching.
                              If ``'equal'``, the value in **matching_criteria** is checked for equality against the
                              stored value.
                              If ``'regex'``, the value in **matching_criteria** is interpreted as a regular expression
                              and is matched against the stored value.
        :type matching_type: str
        :return: Dict mapping error name to :py:obj:`~kafe2.core.error.GaussianErrorBase`-derived error objects.
        :rtype: dict[str, kafe2.core.error.GaussianErrorBase]
        """
        if matching_criteria is not None:
            _crit_ref_value = matching_criteria.pop('reference', None)
            if _crit_ref_value == 'data':
                return self._data_container.get_matching_errors(matching_criteria, matching_type=matching_type)
            if _crit_ref_value == 'model':
                return self._param_model.get_matching_errors(matching_criteria, matching_type=matching_type)
            if _crit_ref_value is None:
                pass  # don't raise, continue evaluation below
            else:
                raise ValueError("Unknown value '{}' for matching "
                                 "criterion 'reference'. Valid: 'data', 'model' or None".format(_crit_ref_value))

        _result = self._data_container.get_matching_errors(matching_criteria, matching_type=matching_type)
        _result_model = self._param_model.get_matching_errors(matching_criteria, matching_type=matching_type)

        # be paranoid about collisions
        for _k in _result_model:
            assert _k not in _result  # FATAL: there is an error with the same name in the data and model containers
            _result[_k] = _result_model[_k]

        return _result

    def add_error(self, err_val, name=None, correlation=0, relative=False, reference='data', **kwargs):
        """Add an uncertainty source to the fit.

        :param err_val: Pointwise uncertainty/uncertainties for all data points.
        :type err_val: float or typing.Iterable[float]
        :param name: Unique name for this uncertainty source. If :py:obj:`None`, the name of the error source will be
                     set to a random alphanumeric string.
        :type name: str or None
        :param correlation: Correlation coefficient between any two distinct data points.
        :type correlation: float
        :param relative: If :py:obj:`True`, **err_val** will be interpreted as a *relative* uncertainty.
        :type relative: bool
        :param reference: Either ``'data'`` or ``'model'``. Specifies which reference values to use when calculating
                          absolute errors from relative errors.
        :type reference: str
        :return: An error id which uniquely identifies the created error source.
        :rtype: str
        """
        if reference == 'data':
            # delegate to data container
            _reference_object = self._data_container
        elif reference == 'model':
            # delegate to model container
            _reference_object = self._param_model
        else:
            raise ValueError("Cannot add error: unknown reference specification '{}',"
                             "expected one of: 'data', 'model'...".format(reference))

        _ret = _reference_object.add_error(err_val=err_val,
                                           name=name, correlation=correlation, relative=relative, **kwargs)

        return _ret

    def add_matrix_error(self, err_matrix, matrix_type,
                         name=None, err_val=None, relative=False, reference='data', **kwargs):
        """Add a matrix uncertainty source for use in the fit.

        :param err_matrix: covariance or correlation matrix
        :param matrix_type: One of ``'covariance'``/``'cov'`` or ``'correlation'``/``'cor'``
        :type matrix_type: str
        :param name: Unique name for this uncertainty source. If :py:obj:`None`, the name of the error source will be
                     set to a random alphanumeric string.
        :type name: str or None
        :param err_val: The pointwise uncertainties (mandatory if only a correlation matrix is given).
        :type err_val: typing.Iterable[float]
        :param relative: If :py:obj:`True`, the covariance matrix and/or **err_val** will be interpreted as a *relative*
                         uncertainty.
        :type relative: bool
        :param reference: Either ``'data'`` or ``'model'``. Specifies which reference values to use when calculating
                          absolute errors from relative errors.
        :type reference: str
        :return: An error id which uniquely identifies the created error source.
        :rtype: str
        """
        if reference == 'data':
            # delegate to data container
            _reference_object = self._data_container
        elif reference == 'model':
            # delegate to model container
            _reference_object = self._param_model
            if relative:
                raise NotImplementedError("Errors relative to model not implemented!")
        else:
            raise ValueError("Cannot add matrix error: unknown reference "
                               "specification '{}', expected one of: 'data', 'model'...".format(reference))

        _ret = _reference_object.add_matrix_error(err_matrix=err_matrix, matrix_type=matrix_type,
                                                  name=name, err_val=err_val, relative=relative, **kwargs)

        return _ret

    def disable_error(self, err_id):
        """Temporarily disable an uncertainty source so that it doesn't count towards calculating the total uncertainty.

        :param str err_id: error id
        """
        try:
            # try to find error in data container
            _ret = self._data_container.disable_error(err_id)  # TODO: this call does not return anything
        except ValueError:
            # try to find error in model container
            _ret = self._param_model.disable_error(err_id)  # TODO: this call does not return anything
        return _ret

    def enable_error(self, err_id):
        """(Re-)Enable an uncertainty source so that it counts towards calculating the total uncertainty.

        :param str err_id: error id
        """
        try:
            # try to find error in data container
            _ret = self._data_container.enable_error(err_id)  # TODO: this call does not return anything
        except ValueError:
            # try to find error in model container
            _ret = self._param_model.enable_error(err_id)  # TODO: this call does not return anything
        return _ret

    def do_fit(self, asymmetric_parameter_errors=False):
        """Perform the minimization of the cost function.

        :param bool asymmetric_parameter_errors: If :py:obj:`True`, calculate asymmetric parameter errors.
        :return: A dictionary containing the fit results.
        :rtype: dict
        """
        if self._cost_function.needs_errors and not self.has_errors:
            warnings.warn("Cost function expects errors but no errors were specified.")
        if self._implicit_no_errors:
            warnings.warn(
                "No data/model errors were specified. Parameter errors cannot be calculated.")

        if self._cost_function_pointwise is not None:
            if is_diagonal(self.total_cov_mat):
                _cost_target = self._cost_function_pointwise.name
            else:
                _cost_target = self._cost_function.name
            self._fitter.parameter_to_minimize = _cost_target

        # Give relative model errors data as reference for initial fit:
        self._set_data_as_model_ref()

        # Initial fit:
        self._pre_fit_iteration(first_fit=True)
        self._fitter.do_fit()  # TODO specify other node to minimize
        self._post_fit_iteration(first_fit=True)

        if self._iterative_fits_needed():
            _convergence_limit = float(kc("fit", "iterative_do_fit", "convergence_limit"))
            _previous_cost = self.cost_function_value
            for i in range(kc("fit", "iterative_do_fit", "max_iterations")):
                self._pre_fit_iteration()
                self._fitter.reset_minimizer()  # flush iminuit cache
                self._fitter.do_fit()
                self._post_fit_iteration()
                if abs(self.cost_function_value - _previous_cost) < _convergence_limit:
                    break
                _previous_cost = self.cost_function_value
        elif self._second_fit_needed():
            self._pre_fit_iteration()
            self._fitter.reset_minimizer()  # flush iminuit cache
            self._fitter.do_fit()
            self._post_fit_iteration()

        self._loaded_result_dict = None
        self._update_parameter_formatters()
        return self.get_result_dict(asymmetric_parameter_errors=asymmetric_parameter_errors)

    def assign_model_function_name(self, name):
        """Assign a string to be the model function name.

        :param str name: The new name.
        """
        self._model_function.formatter.name = name

    def assign_model_function_expression(self, expression_format_string):
        """Assign a plain-text-formatted expression string to the model function.

        :param str expression_format_string: The plain text string.
        """
        self._model_function.formatter.expression_format_string = expression_format_string

    def assign_parameter_names(self, **par_names_dict):
        """Assign display strings to all model function arguments.

        :param par_names_dict: Dictionary mapping the parameter names to their display names.
        """
        for _af in self._get_model_function_argument_formatters():
            _an = par_names_dict.pop(_af.arg_name, None)
            if _an is not None:
                _af.name = _an
        if par_names_dict:
            warnings.warn("Could not assign all names to a parameter."
                          "Leftover: {}".format(par_names_dict))

    def assign_model_function_latex_name(self, latex_name):
        """Assign a LaTeX-formatted string to be the model function name.

        :param str latex_name: The LaTeX string.
        """
        self._model_function.formatter.latex_name = latex_name

    def assign_model_function_latex_expression(self, latex_expression_format_string):
        """Assign a LaTeX-formatted expression string to the model function.

        :param str latex_expression_format_string: The LaTeX string. Elements like ``'{par_name}'``
            will be replaced automatically with the corresponding LaTeX names for the given
            parameter. These can be set with :py:meth:`~assign_parameter_latex_names`.
        """
        self._model_function.formatter.latex_expression_format_string = \
            latex_expression_format_string

    def assign_parameter_latex_names(self, **par_latex_names_dict):
        """Assign LaTeX-formatted strings to all model function arguments.

        :param par_latex_names_dict: Dictionary mapping the parameter names to their latex names.
        """
        for _af in self._get_model_function_argument_formatters():
            _aln = par_latex_names_dict.pop(_af.arg_name, None)
            if _aln is not None:
                _af.latex_name = _aln
        if par_latex_names_dict:
            warnings.warn("Could not assign all latex names to a parameter."
                          "Leftover: {}".format(par_latex_names_dict))

    def get_result_dict(self, asymmetric_parameter_errors=False):
        """Return a dictionary of the fit results.

        :param bool asymmetric_parameter_errors: If :py:obj:`True`, calculate asymmetric parameter errors.
        :return: A dictionary containing the fit results.
        :rtype: dict
        """
        _result_dict = dict()

        _result_dict['did_fit'] = self.did_fit
        _cost = float(self.cost_function_value)  # convert numpy scalar to float for yaml representation
        _ndf = self.ndf
        _result_dict['cost'] = _cost
        _result_dict['ndf'] = _ndf
        _gof = self.goodness_of_fit
        _result_dict['goodness_of_fit'] = _gof
        _result_dict['gof/ndf'] = _gof / _ndf if _gof is not None else _gof
        _result_dict['chi2_probability'] = self.chi2_probability
        _result_dict['parameter_values'] = self.parameter_name_value_dict
        if _result_dict['did_fit']:
            _result_dict['parameter_cov_mat'] = self.parameter_cov_mat
            _parameter_errors = OrderedDict()
            for _pn, _pe in zip(self.parameter_names, self.parameter_errors):
                _parameter_errors[_pn] = _pe
            _result_dict['parameter_errors'] = _parameter_errors
            _result_dict['parameter_cor_mat'] = self.parameter_cor_mat
        else:
            _result_dict['parameter_cov_mat'] = None
            _result_dict['parameter_errors'] = None
            _result_dict['parameter_cor_mat'] = None

        if self._loaded_result_dict is not None and self._loaded_result_dict['asymmetric_parameter_errors'] is not None:
            _asymm_errs = self._loaded_result_dict['asymmetric_parameter_errors']
        elif asymmetric_parameter_errors:
            self._check_dynamic_error_compatibility()
            _asymm_errs = self.asymmetric_parameter_errors
        else:
            _asymm_errs = self._fitter.asymmetric_fit_parameter_errors_if_calculated
        if _asymm_errs is not None:
            _asymm_errs_dict = OrderedDict()
            for _pn, _ape in zip(self.parameter_names, _asymm_errs):
                _asymm_errs_dict[_pn] = _ape
            _asymm_errs = _asymm_errs_dict
        _result_dict['asymmetric_parameter_errors'] = _asymm_errs

        return _result_dict

    def report(self, output_stream=sys.stdout, show_data=True, show_model=True, show_fit_results=True,
               asymmetric_parameter_errors=False):
        """Print a summary of the fit state and/or results.

        :param output_stream: The output stream to which the report should be printed.
        :type output_stream: io.TextIOBase
        :param show_data: If :py:obj:`True`, print out information about the data.
        :type show_data: bool
        :param show_model: If :py:obj:`True`, print out information about the parametric model.
        :type show_model: bool
        :param show_fit_results: If :py:obj:`True`, print out information about the fit results.
        :type show_fit_results: bool
        :param asymmetric_parameter_errors: If :py:obj:`True`, use two different parameter errors for up/down
            directions.
        :type asymmetric_parameter_errors: bool
        """

        _indent = ' ' * 4
        if show_data:
            self._report_data(output_stream=output_stream, indent=_indent, indentation_level=0)
        if show_model:
            self._report_model(output_stream=output_stream, indent=_indent, indentation_level=0)
        if show_fit_results:
            self._report_fit_results(output_stream=output_stream, indent=_indent, indentation_level=0,
                                     asymmetric_parameter_errors=asymmetric_parameter_errors)

    def to_file(self, filename, file_format=None, calculate_asymmetric_errors=False):
        """Write kafe2 object to file

        :param filename: Filename for the output.
        :type filename: str
        :param file_format: A format for the output file. If :py:obj:`None`, the extension from the filename is used.
        :type file_format: str or None
        :param calculate_asymmetric_errors: If asymmetric errors should be calculated before saving the results.
        :type calculate_asymmetric_errors: bool
        """
        if calculate_asymmetric_errors:
            _ = self.asymmetric_parameter_errors  # force calculation of asymmetric errors if not calculated yet
        super(FitBase, self).to_file(filename=filename, file_format=file_format)

    def save_state(
            self, filename: str, file_format: Optional[str] = None,
            calculate_asymmetric_errors: bool = False):
        """Write current state of the fit to file.
        Unlike to `to_file` this does not contain information regarding how the fit is constructed -
        this is because for complex fit objects a reconstruction from e.g. YAML may not work.
        In such cases the fit object should be contructed via Python and `save_state` and
        `load_state` should be used.

        :param filename: Filename for the output.
        :type filename: str
        :param file_format: A format for the output file. If :py:obj:`None`, the extension from the
            filename is used.
        :type file_format: str or None
        :param calculate_asymmetric_errors: If asymmetric errors should be calculated before saving
            the results.
        :type calculate_asymmetric_errors: bool
        """
        if file_format is None:
            _, _extension = os.path.splitext(filename)
            if len(_extension) >= 2:
                file_format = _extension[1:]
        if file_format is None:
            file_format = "yml"
        file_format = file_format.lower()

        if calculate_asymmetric_errors:
            _ = self.asymmetric_parameter_errors

        if file_format in ("yaml", "yml"):
            with open(filename, "w", encoding="utf8") as _f:
                yaml.safe_dump(to_python_types(self.get_result_dict()), _f)
        else:
            raise ValueError(f"Unknown file format: {file_format}. Available: yaml")

    def load_state(self, filename: str, file_format: Optional[str] = None):
        """load current fit state from the specified file that was written with `save_state`.

        :param filename: Filename for the input.
        :type filename: str
        :param file_format: A format for the output file. If :py:obj:`None`, the extension from the
            filename is used.
        :type file_format: str or None
        """
        if file_format is None:
            _, _extension = os.path.splitext(filename)
            if len(_extension) >= 2:
                file_format = _extension[1:]
        if file_format is None:
            file_format = "yml"
        file_format = file_format.lower()

        if file_format in ("yaml", "yml"):
            with open(filename, "r", encoding="utf8") as _f:
                self._loaded_result_dict = to_numpy_arrays(yaml.safe_load(_f))
        else:
            raise ValueError(f"Unknown file format: {file_format}. Available: yaml")
        _new_par_vals = self._loaded_result_dict.pop("parameter_values", None)
        if _new_par_vals is not None:
            self._fitter.set_all_fit_parameter_values(_new_par_vals)



import unittest
import numpy as np
import scipy.stats as stats

from kafe2.fit import HistContainer, HistParametricModel
from kafe2.fit.histogram.model import HistModelFunction
from kafe2.fit.util.function_library import linear_model, quadratic_model


class TestDatastoreHistogram(unittest.TestCase):

    def setUp(self):
        self._ref_entries = [-9999., -8279., 3.3, 5.5, 2.2, 8.5, 10., 10.2, 10000., 1e7]
        self._ref_n_bins_auto = 25
        self._ref_n_bins_manual = 10
        self._ref_n_bin_range = (0., 10.)

        self._ref_bin_edges_manual_equalspacing = np.linspace(0, 10, self._ref_n_bins_manual + 1)
                                                    # [0 , 1 , 2 , 3 , 4 , 5 , 6 , 7 , 8 , 9 , 10]
        self._ref_data_manual_equalspacing = np.array([  0,  0,  1,  1,  0,  1,  0,  0,  1,  0])
        self._ref_bin_heights_manual =       np.array([  5,  4, 10,  3,  0, 15,  7,  4, 20,  1])

        self._ref_bin_edges_manual_variablespacing =   [0 , 2 , 3 , 3.1 , 3.2 , 3.3 , 3.4 , 7 , 8.5 , 9 , 10]
        self._ref_data_manual_variablespacing = np.array([  0,  1,  0,    0,    0,    1,    1,  0,    1,  0])

        self._ref_data_auto = np.array([0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0])

        # test rebinning functionality
        self._probe_bin_edges_variablespacing_withedges =    [0,  2, 3, 3.1, 3.2, 3.3, 3.4, 7, 8.5, 9, 10]    # OK
        self._probe_bin_edges_variablespacing_noedges =      [    2, 3, 3.1, 3.2, 3.3, 3.4, 7, 8.5, 9]        # OK
        self._probe_bin_edges_variablespacing_wrongedges1 =  [0,  2, 3, 3.1, 3.2, 3.3, 3.4, 7, 8.5, 9, 12.3]  # fail
        self._probe_bin_edges_variablespacing_wrongedges2 =  [-9, 2, 3, 3.1, 3.2, 3.3, 3.4, 7, 8.5, 9, 10]   # fail
        self._probe_bin_edges_variablespacing_wrongedges3 =  [-3, 2, 3, 3.1, 3.2, 3.3, 3.4, 7, 8.5, 9, 22]   # fail
        self._probe_bin_edges_variablespacing_wrongnumber =  [0,  2, 3, 3.1, 3.2, 3.3, 3.4, 7, 8.5, 10]       # fail
        self._probe_bin_edges_variablespacing_unsorted =     [0,  2, 3, 8.5, 3.2, 3.3, 3.4, 7, 3.1, 9, 10]    # fail


        self.hist_cont_binedges_auto = HistContainer(self._ref_n_bins_auto, self._ref_n_bin_range, bin_edges=None)
        self.hist_cont_binedges_manual_equal = HistContainer(self._ref_n_bins_manual, self._ref_n_bin_range, bin_edges=self._ref_bin_edges_manual_equalspacing)
        self.hist_cont_binedges_manual_variable = HistContainer(self._ref_n_bins_manual, self._ref_n_bin_range, bin_edges=self._ref_bin_edges_manual_variablespacing)


    def test_property_size(self):
        self.assertEqual(self.hist_cont_binedges_auto.size, self._ref_n_bins_auto)
        self.assertEqual(self.hist_cont_binedges_manual_equal.size, self._ref_n_bins_manual)
        self.assertEqual(self.hist_cont_binedges_manual_variable.size, self._ref_n_bins_manual)

    def test_property_low(self):
        self.assertEqual(self.hist_cont_binedges_auto.low, self._ref_n_bin_range[0])
        self.assertEqual(self.hist_cont_binedges_manual_equal.low, self._ref_n_bin_range[0])
        self.assertEqual(self.hist_cont_binedges_manual_variable.low, self._ref_n_bin_range[0])

    def test_property_high(self):
        self.assertEqual(self.hist_cont_binedges_auto.high, self._ref_n_bin_range[1])
        self.assertEqual(self.hist_cont_binedges_manual_equal.high, self._ref_n_bin_range[1])
        self.assertEqual(self.hist_cont_binedges_manual_variable.high, self._ref_n_bin_range[1])

    def test_fill_empty_binedges_auto_compare_data(self):
        self.hist_cont_binedges_auto.fill(self._ref_entries)
        self.assertTrue(
            np.allclose(self.hist_cont_binedges_auto.data, self._ref_data_auto)
        )

    def test_fill_empty_binedges_manual_equal_compare_data(self):
        self.hist_cont_binedges_manual_equal.fill(self._ref_entries)
        self.assertTrue(
            np.allclose(self.hist_cont_binedges_manual_equal.data, self._ref_data_manual_equalspacing)
        )

    def test_fill_empty_binedges_manual_variable_compare_data(self):
        self.hist_cont_binedges_manual_variable.fill(self._ref_entries)
        self.assertTrue(
            np.allclose(self.hist_cont_binedges_manual_variable.data, self._ref_data_manual_variablespacing)
        )

    def test_fill_empty_binedges_auto_rebin_manual_equal_compare_data(self):
        self.hist_cont_binedges_auto.fill(self._ref_entries)
        self.hist_cont_binedges_auto.rebin(self._ref_bin_edges_manual_equalspacing)
        self.assertTrue(
            np.allclose(self.hist_cont_binedges_auto.data, self._ref_data_manual_equalspacing)
        )

    def test_fill_empty_binedges_auto_rebin_manual_variable_compare_data(self):
        self.hist_cont_binedges_auto.fill(self._ref_entries)
        self.hist_cont_binedges_auto.rebin(self._ref_bin_edges_manual_variablespacing)
        self.assertTrue(
            np.allclose(self.hist_cont_binedges_auto.data, self._ref_data_manual_variablespacing)
        )

    def test_manual_bin_height(self):
        self.hist_cont_binedges_manual_equal.set_bins(self._ref_bin_heights_manual)
        self.assertTrue(np.alltrue(self.hist_cont_binedges_manual_equal.data == self._ref_bin_heights_manual))
        self.assertTrue(self.hist_cont_binedges_manual_equal._manual_heights)

    def test_construct_bin_edges_variablespacing_withedges(self):
        _hc = HistContainer(self._ref_n_bins_manual, self._ref_n_bin_range, bin_edges=self._probe_bin_edges_variablespacing_withedges)

    def test_construct_bin_edges_variablespacing_noedges(self):
        _hc = HistContainer(self._ref_n_bins_manual, self._ref_n_bin_range, bin_edges=self._probe_bin_edges_variablespacing_noedges)

    def test_raise_construct_bin_edges_variablespacing_wrongedges1(self):
        with self.assertRaises(ValueError):
            _hc = HistContainer(self._ref_n_bins_manual, self._ref_n_bin_range,
                                bin_edges=self._probe_bin_edges_variablespacing_wrongedges1)

    def test_raise_construct_bin_edges_variablespacing_wrongedges2(self):
        with self.assertRaises(ValueError):
            _hc = HistContainer(self._ref_n_bins_manual, self._ref_n_bin_range,
                                bin_edges=self._probe_bin_edges_variablespacing_wrongedges2)

    def test_raise_construct_bin_edges_variablespacing_wrongedges3(self):
        with self.assertRaises(ValueError):
            _hc = HistContainer(self._ref_n_bins_manual, self._ref_n_bin_range,
                                bin_edges=self._probe_bin_edges_variablespacing_wrongedges3)

    def test_raise_construct_bin_edges_variablespacing_wrongnumber(self):
        with self.assertRaises(ValueError):
            _hc = HistContainer(self._ref_n_bins_manual, self._ref_n_bin_range,
                                bin_edges=self._probe_bin_edges_variablespacing_wrongnumber)

    def test_raise_construct_bin_edges_variablespacing_unsorted(self):
        with self.assertRaises(ValueError):
            _hc = HistContainer(self._ref_n_bins_manual, self._ref_n_bin_range,
                                bin_edges=self._probe_bin_edges_variablespacing_unsorted)

    def test_raise_add_same_error_name_twice(self):
        self.hist_cont_binedges_auto.add_error(0.1,
                                               name="MyNewError",
                                               correlation=0, relative=False)
        with self.assertRaises(ValueError):
            self.hist_cont_binedges_auto.add_error(0.1,
                                                   name="MyNewError",
                                                   correlation=0, relative=False)

    def test_raise_get_inexistent_error(self):
        with self.assertRaises(ValueError):
            self.hist_cont_binedges_auto.get_error("MyInexistentError")

    def test_raise_manual_bin_heights_rebin(self):
        self.hist_cont_binedges_manual_equal.set_bins(self._ref_bin_heights_manual)
        with self.assertRaises(RuntimeError):
            self.hist_cont_binedges_manual_equal.rebin(self._ref_bin_edges_manual_variablespacing)

    def test_raise_add_error_bad_shape(self):
        with self.assertRaises(ValueError):
            self.hist_cont_binedges_auto.add_error([1, 2, 3])

    def test_raise_add_matrix_error_bad_shape(self):
        with self.assertRaises(ValueError):
            self.hist_cont_binedges_auto.add_matrix_error([[1, 0], [0, 4]], "cov")
        with self.assertRaises(ValueError):
            self.hist_cont_binedges_auto.add_matrix_error(np.eye(5), "cor", err_val=np.ones(10))
        with self.assertRaises(ValueError):
            self.hist_cont_binedges_auto.add_matrix_error(np.eye(5), "cor", err_val=np.eye(5))

    def test_raise_bad_data_shape(self):
        with self.assertRaises(ValueError):
            self.hist_cont_binedges_auto.fill([[1, 2], [3, 4]])
        with self.assertRaises(ValueError):
            self.hist_cont_binedges_auto.set_bins(np.eye(self._ref_n_bins_auto))

    def test_warn_int_data(self):
        container = HistContainer(n_bins=10, bin_range=(0.0, 10.0), fill_data=[1, 2, 3])
        with self.assertWarns(Warning):
            _ = container.data

class TestDatastoreHistParametricModel(unittest.TestCase):
    @staticmethod
    def _ref_model_func(x, mu, sigma):
        return stats.norm(mu, sigma).pdf(x)

    @staticmethod
    def _ref_model_func_antider(x, mu, sigma):
        return stats.norm(mu, sigma).cdf(x)

    def setUp(self):
        self._ref_pm_support = np.linspace(-5, 5, 11)

        self._ref_n_bins = 11
        self._ref_n_bin_range = (-3, 25)
        self._ref_bin_edges = np.linspace(self._ref_n_bin_range[0], self._ref_n_bin_range[1], self._ref_n_bins+1)

        self._ref_params = (14., 3.)
        self._ref_data = (self._ref_model_func_antider(self._ref_bin_edges[1:], *self._ref_params) -
                          self._ref_model_func_antider(self._ref_bin_edges[:-1], *self._ref_params))

        self.hist_param_model_numerical = HistParametricModel(
            n_bins=self._ref_n_bins,
            bin_range=self._ref_n_bin_range,
            model_density_func=HistModelFunction(self._ref_model_func), 
            model_parameters=self._ref_params,
            bin_edges=None, 
            bin_evaluation="numerical")

        self.hist_param_model_with_antider = HistParametricModel(
            n_bins=self._ref_n_bins,
            bin_range=self._ref_n_bin_range,
            model_density_func=HistModelFunction(self._ref_model_func),
            model_parameters=self._ref_params,
            bin_edges=None, bin_evaluation=self._ref_model_func_antider)

        self.hist_param_model_np_vectorize = HistParametricModel(
            n_bins=self._ref_n_bins, bin_range=self._ref_n_bin_range,
            model_density_func=self._ref_model_func,
            model_parameters=self._ref_params, bin_edges=None,
            bin_evaluation=np.vectorize(self._ref_model_func_antider)
        )

        self._test_params = (20., 5.)
        self._ref_test_data = (self._ref_model_func_antider(self._ref_bin_edges[1:], *self._test_params) -
                          self._ref_model_func_antider(self._ref_bin_edges[:-1], *self._test_params))

    def test_compare_hist_model_no_antider_ref_data(self):
        self.assertTrue(np.allclose(self.hist_param_model_numerical.data, self._ref_data))

    def test_compare_hist_model_with_antider_ref_data(self):
        self.assertTrue(np.allclose(self.hist_param_model_with_antider.data, self._ref_data))

    def test_compare_hist_model_np_vectorize_ref_data(self):
        self.assertTrue(np.allclose(self.hist_param_model_np_vectorize.data, self._ref_data))

    def test_change_parameters_test_data(self):
        self.hist_param_model_numerical.parameters = self._test_params
        self.assertTrue(np.allclose(self.hist_param_model_numerical.data, self._ref_test_data))

    def test_raise_set_data(self):
        with self.assertRaises(TypeError):
            self.hist_param_model_numerical.data = self._ref_test_data

    def test_raise_fill(self):
        with self.assertRaises(TypeError):
            self.hist_param_model_numerical.fill([-1, 2, 700])

    def test_raise_unknown_bin_evaluation_raise(self):
        with self.assertRaises(ValueError):
            HistParametricModel(n_bins=10, bin_range=(-5, 5), bin_evaluation="magical")
        with self.assertRaises(ValueError):
            HistParametricModel(n_bins=10, bin_range=(-5, 5), bin_evaluation=5)


class TestQuadrature(unittest.TestCase):

    @staticmethod
    def _get_models(bin_evaluation):

        def _sinus_model(x, a):
            return np.sin(x) + a

        return (
            HistParametricModel(
                n_bins=10, bin_range=(0, 10), model_density_func=linear_model,
                model_parameters=[1.0, 1.0], bin_evaluation=bin_evaluation
            ), HistParametricModel(
                n_bins=10, bin_range=(0, 10), model_density_func=quadratic_model,
                model_parameters=[-0.1, 1.0, 1.0], bin_evaluation=bin_evaluation
            ), HistParametricModel(
                n_bins=1000, bin_range=(0, 10), model_density_func=quadratic_model,
                model_parameters=[-0.1, 1.0, 1.0], bin_evaluation=bin_evaluation
            ), HistParametricModel(
                n_bins=1000, bin_range=(0, 2 * np.pi), model_density_func=_sinus_model,
                model_parameters=[2.0], bin_evaluation=bin_evaluation
            )
        )

    def setUp(self):
        (
            self._linear_model_numerical,
            self._quadratic_model_numerical,
            self._quadratic_model_limit_numerical,
            self._sinus_model_limit_numerical
        ) = TestQuadrature._get_models("numerical")

    def test_rectangle(self):
        (
            _linear_model_rectangle,
            _quadratic_model_rectangle,
            _quadratic_model_limit_rectangle,
            _sinus_model_limit_rectangle
        ) = TestQuadrature._get_models("rectangle")
        self.assertTrue(np.allclose(
            self._linear_model_numerical.data, _linear_model_rectangle.data
        ))
        self.assertTrue(np.allclose(
            self._quadratic_model_limit_numerical.data, _quadratic_model_limit_rectangle.data
        ))
        self.assertTrue(np.allclose(
            self._sinus_model_limit_numerical.data, _sinus_model_limit_rectangle.data
        ))

    def test_trapezoid(self):
        (
            _linear_model_trapezoid,
            _quadratic_model_trapezoid,
            _quadratic_model_limit_trapezoid,
            _sinus_model_limit_rectangle
        ) = TestQuadrature._get_models("trapezoid")
        self.assertTrue(np.allclose(
            self._linear_model_numerical.data, _linear_model_trapezoid.data
        ))
        self.assertTrue(np.allclose(
            self._quadratic_model_limit_numerical.data, _quadratic_model_limit_trapezoid.data
        ))
        self.assertTrue(np.allclose(
            self._sinus_model_limit_numerical.data, _sinus_model_limit_rectangle.data
        ))

    def test_simpson(self):
        (
            _linear_model_simpson,
            _quadratic_model_simpson,
            _quadratic_model_limit_simpson,
            _sinus_model_limit_simpson
        ) = TestQuadrature._get_models("simpson")
        self.assertTrue(np.allclose(
            self._linear_model_numerical.data, _linear_model_simpson.data
        ))
        self.assertTrue(np.allclose(
            self._quadratic_model_numerical.data, _quadratic_model_simpson.data
        ))
        self.assertTrue(np.allclose(
            self._quadratic_model_limit_numerical.data, _quadratic_model_limit_simpson.data
        ))
        self.assertTrue(np.allclose(
            self._sinus_model_limit_numerical.data, _sinus_model_limit_simpson.data
        ))



from .._base.cost import CostFunction, CostFunction_Chi2, CostFunction_NegLogLikelihood, \
    CostFunction_GaussApproximation


__all__ = [
    "HistCostFunction",
    "HistCostFunction_Chi2",
    "HistCostFunction_NegLogLikelihood",
    "HistCostFunction_GaussApproximation"
]


class HistCostFunction(CostFunction):
    pass


class HistCostFunction_Chi2(CostFunction_Chi2):
    pass


class HistCostFunction_NegLogLikelihood(CostFunction_NegLogLikelihood):
    pass


class HistCostFunction_GaussApproximation(CostFunction_GaussApproximation):
    pass



from ..minimizers import MinimizerIMinuit


class SimpleFitter(object):
    def __init__(self, nexus, parameters_to_fit, parameter_to_minimize, minimizer_class=MinimizerIMinuit):
        self._nx = nexus
        self.parameters_to_fit = parameters_to_fit
        self.parameter_to_minimize = parameter_to_minimize

        # local cache
        self.__cache_stale = True
        self.__cache_all_parameters_name_value_dict = None
        self.__cache_fit_parameters_name_value_dict = None
        self.__cache_parameter_to_minimize_value = None

        _par_name_val_map = self.fit_parameter_values
        self._minimizer = minimizer_class(parameters_to_fit,
                                          _par_name_val_map.values(),
                                          [0.1 if _v==0 else 0.1*_v for _v in _par_name_val_map.values()],
                                          self._fcn_wrapper)

    # -- private methods

    def _check_parnames_in_par_space_raise(self, fit_pars):
        for _pn in fit_pars:
            if self._nx.get_by_name(_pn) is None:
                raise ValueError("No parameter with name '%s' registered in ParameterSpace (%r)!"
                                 % (_pn, self._nx))

    def _renew_par_cache(self):
        _fpns = self.parameters_to_fit
        _apnvd = self.__cache_all_parameters_name_value_dict = self._nx.parameter_values_dict
        self.__cache_fit_parameters_name_value_dict = OrderedDict([(_pn, _apnvd[_pn]) for _pn in _fpns])
        self.__cache_parameter_to_minimize_value = _apnvd[self.parameter_to_minimize]

    def _minimize(self, max_calls=6000):
        self._minimizer.minimize(max_calls=max_calls)
        # opt.minimize(self._fcn_wrapper,
        #              self.fit_parameter_values.values(),
        #              args=(), method=None, jac=None,
        #              hess=None, hessp=None,
        #              bounds=None, constraints=(),
        #              tol=None,
        #              callback=None,
        #              options=None)
        self.__cache_stale = True


    def _fcn_wrapper(self, *fit_par_value_list):
        # # set parameters to current values
        # if not (len(fit_par_value_list) == 1 and fit_par_value_list[0] is None):
        #     if not len(fit_par_value_list) == self.n_fit_par:
        #         print fit_par_value_list
        #         print self.n_fit_par
        # set parameters to current values
        _par_val_dict = {_pn: _pv for _pn, _pv in zip(self.parameters_to_fit, fit_par_value_list)}
        self._nx.set(**_par_val_dict)
        # evaluate function and return value
        return self._nx.get_values_by_name(self.parameter_to_minimize)

    # -- public properties

    @property
    def minimizer(self):
        return self._minimizer

    @property
    def parameters_to_fit(self):
        return self._fit_pars

    @parameters_to_fit.setter
    def parameters_to_fit(self, fit_parameters):
        self._check_parnames_in_par_space_raise(fit_parameters)
        self._fit_pars = fit_parameters

    @property
    def parameter_to_minimize(self):
        return self._parameter_to_minimize

    @parameter_to_minimize.setter
    def parameter_to_minimize(self, parameter_to_minimize):
        self._check_parnames_in_par_space_raise((parameter_to_minimize,))
        self._parameter_to_minimize = parameter_to_minimize

    @property
    def fit_parameter_values(self):
        if self.__cache_stale:
            self._renew_par_cache()
        return self.__cache_fit_parameters_name_value_dict

    @property
    def parameter_to_minimize_value(self):
        if self.__cache_stale:
            self._renew_par_cache()
        return self.__cache_parameter_to_minimize_value

    @property
    def n_fit_par(self):
        return len(self.parameters_to_fit)

    # -- public methods

    def do_fit(self):
        self._minimize()



import abc
import numpy as np
import operator
import six
import sys
import uuid
import logging
import warnings
import weakref

from ast import parse

if six.PY2:
    from funcsigs import signature, Parameter as SigParameter
else:
    from inspect import signature, Parameter as SigParameter

__all__ = ['Nexus', 'Alias', 'Array', 'Fallback', 'Function', 'Parameter', 'Tuple']

# -- helpers

_OPERATORS = {
    # arithmetic
    'add': operator.add,
    'sub': operator.sub,
    'mul': operator.mul,
    'truediv': operator.truediv,
    'floordiv': operator.floordiv,
    'mod': operator.mod,
}

# 'div' operator only available in Python 2
if six.PY2:
    _OPERATORS['div'] = operator.div

_OPERATORS = {
    # arithmetic
    'add': (operator.add, True),
    'sub': (operator.sub, True),
    'mul': (operator.mul, True),
    'truediv': (operator.truediv, True),
    'floordiv': (operator.floordiv, True),
    'mod': (operator.mod, True),
    # binary arithmetic
    'lshift': (operator.lshift, True),
    'rshift': (operator.rshift, True),
    # logical
    'and': (operator.and_, True),
    'or': (operator.or_, True),
    'xor': (operator.xor, True),
    # comparison
    #'eq': (operator.eq, False),
    #'ne': (operator.ne, False),
    #'lt': (operator.lt, False),
    #'le': (operator.le, False),
    #'gt': (operator.gt, False),
    #'ge': (operator.ge, False),
    # other
    #'getitem': (operator.getitem, False),
    #'contains': (operator.contains, False),
}

_UNARY_OPERATORS = {
    'invert': operator.invert,
    'neg': operator.neg,
    'pos': operator.pos,
    'len': len,
}

# 'div' operator only available in Python 2
if six.PY2:
    _OPERATORS['div'] = (operator.div, True)

# '_abs' operator only available in Python 3
if six.PY3:
    _UNARY_OPERATORS['abs'] = operator._abs


def _add_common_operators(cls):
    """class decorator for implementing common operator methods for nodes"""

    for name, (op, add_reversed) in list(_OPERATORS.items()):
        setattr(cls, '__{}__'.format(name), cls._make_binop_method(op))
        if add_reversed:
            setattr(cls, '__r{}__'.format(name), cls._make_rbinop_method(op))
    for name, op in _UNARY_OPERATORS.items():
        setattr(cls, '__{}__'.format(name), cls._make_unaryop_method(op))

    return cls


# -- Nodes -------------------------------------

@six.add_metaclass(abc.ABCMeta)
class NodeBase(object):
    """
    Abstract class for a computation graph node.
    Defines the minimal interface required by all specializations.
    Dependencies between nodes are expressed via parent nodes and child nodes.
    The status of a parent node depends on the status of its children; When the status of a child
    node changes the status of any of its parents will need to be changed as well.
    """

    RESERVED_PARAMETER_NAMES = ('__all__', '__real__', '__root__', '__error__')

    def __init__(self, name=None):
        """
        :param name: the name of the node
        :type name: str
        """
        self.name = name or 'Node_' + uuid.uuid4().hex[:10]

        self._stale = True
        self._frozen = False
        self._children = []
        self._parents = set()

    def __iter__(self):
        raise TypeError("'{}' is not iterable".format(self.__class__.__name__))

    def __str__(self):
        return "{}('{}')  [{}]".format(
            self.__class__.__name__,
            self.name, hex(id(self))
        )

    def __repr__(self):
        return self.__str__()

    def _pprint(self):
        """return pretty-printed version of node"""
        return "{}('{}')".format(
            self.__class__.__name__,
            self.name
        )

    def __eq__(self, other):
        return self is other

    def __ne__(self, other):
        return self is not other

    def __hash__(self):
        return id(self)

    @staticmethod
    def _check_name_raise(name):
        """
        Check if node name is valid.
        :param name: the node name
        :type name: str
        """
        if name in Parameter.RESERVED_PARAMETER_NAMES:
            raise ValueError("Invalid node name: '%s' is a reserved keyword!" % (name,))
        try:
            parse('dict(%s=0.123)' % (name,))
        except SyntaxError:
            raise ValueError("Invalid node name '%s'. Must be Python identifier!" % (name,))

    # -- properties

    @property
    def stale(self):
        """
        :return: if the node needs to be updated before use.
        :rtype: bool
        """
        return self._stale

    @property
    def frozen(self):
        """
        If a node is frozen it will be unaffected by updates to its children. It will also not
        notify its parents of any updates.
        :return: if the node is frozen.
        :rtype: bool
        """
        return self._frozen

    @property
    def name(self):
        """
        :return: the name of the node.
        :rtype: str
        """
        return self._name

    @name.setter
    def name(self, name):
        NodeBase._check_name_raise(name)
        self._name = name

    # -- public API

    def add_child(self, node):
        """
        Add a child node to this node. This node will automatically add itself as a parent of the
        child node.
        :param node: the child node to be added.
        :type node: NodeBase
        """
        if not isinstance(node, NodeBase):
            # wrap non-node values inside `Parameter`
            node = Parameter(node)

        self._children.append(node)
        node.add_parent(self)
        self.mark_for_update()

    def add_parent(self, node):
        """
        Add a parent node to this node. Not to be called manually but as an automatic response to
        NodeBase.add_child!
        :param node: the parent node to be added.
        :type node: NodeBase
        """
        if not isinstance(node, NodeBase):
            raise TypeError(
                "Cannot add parent: expected "
                "node type, got {}".format(type(node))
            )
        if self not in node.get_children():
            raise ValueError("Child must be added to parent, not the other way around!")
        self._parents.add(weakref.ref(node))

    def get_children(self):
        """
        :return: a list containing the children of this node.
        :rtype: list of NodeBase
        """
        return self._children

    def get_parents(self):
        """
        :return: a list containing the parents of this node.
        :rtype: list of NodeBase
        """
        return list(self.iter_parents())

    def remove_child(self, node):
        """
        Remove a child node from this node. Automatically removes this node from parents of child
        node.
        :param node: the child node to be removed.
        :type node: NodeBase
        """
        if not isinstance(node, NodeBase):
            raise TypeError(
                "Cannot remove child: expected "
                "node type, got {}".format(type(node))
            )
        # remove all instances of the child node from root children
        self._children = [
            _child
            for _child in self._children
            if _child is not node
        ]

        # remove self node from node parents
        node.remove_parent(self)
        self.mark_for_update()

    def remove_parent(self, node):
        """
        Remove a parent node from this node. Not to be called manually but as an automatic response
        to NodeBase.remove_child!
        :param node: the parent node to be removed.
        :type node: NodeBase
        """
        if not isinstance(node, NodeBase):
            raise TypeError(
                "Cannot remove parent: expected "
                "node type, got {}".format(type(node))
            )

        if self in node.get_children():
            raise ValueError("Must remove child from parent, not the other way around!")

        # remove node from node parents
        self._parents.remove(weakref.ref(node))

        # do *not* remove self node from node children

    def iter_parents(self):
        """
        Generator for parent nodes.
        """
        _out_of_scope_refs = []
        for _p_ref in self._parents:
            _p = _p_ref()  # dereference weakref

            # parent went out of scope
            if _p is None:
                _out_of_scope_refs.append(_p_ref)
                continue

            yield _p

        # cleanup
        if _out_of_scope_refs:
            logging.debug(
                "{} parent refs went out of scope since last "
                "call to 'iter_parents' and will be removed.".format(
                    len(_out_of_scope_refs)
                )
            )
        for _ref in _out_of_scope_refs:
            self._parents.remove(_ref)

    def set_children(self, children):
        """
        Setter method for children. New children that are not of type NodeBase will be wrapped in
        a Parameter node.
        :param children: the new children of this node.
        :type children: iterable
        """
        _new_children = []
        for _child in children:
            # wrap literal values inside 'Parameter'
            if not isinstance(_child, NodeBase):
                _child = Parameter(_child)

            _new_children.append(_child)

        self._children = _new_children

        for _child in self._children:
            _child.add_parent(self)

        self.mark_for_update()

    def notify_parents(self):
        """
        Notify parents that they will need to be updated because this node has changed.
        """
        for _p in self.iter_parents():
            _p.mark_for_update()

    def freeze(self):
        """
        Sets this node's frozen property to True.
        """
        self._frozen = True

    def unfreeze(self):
        """
        Sets this node's frozen property to False.
        """
        self._frozen = False
        self._stale = True

    def mark_for_update(self):
        """
        Sets this node's stale property to True.
        """
        if not self.stale and not self.frozen:
            self._stale = True
            self.notify_parents()

    def update(self):
        """
        Sets this node's stale property to False and performs any necessary updates to this node.
        """
        self._stale = False

    def replace(self, other, other_children=True):
        """
        Replace all instances of this node with `other`. The order of children will be preserved.
        :arg other: the node that this node should be replaced with.
        :type other: any
        :arg other_children: if True, `other` will keep its own children after the replacement.
        If False, only `other` will be replaced: after the replacement other will have the same
        children as this node.
        :type other_children: bool
        """
        # Do nothing if a node is supposed to be replaced with itself
        if self is other:
            return

        if not isinstance(other, NodeBase):
            other = Parameter(other)

        if not other_children:
            other.set_children(self.get_children())

        for _parent in self.get_parents():
            _parent.replace_child(current_child=self, new_child=other)

    def replace_child(self, current_child, new_child):
        """
        Replace a child node of this node with another node. If new_child is not of type NodeBase
        it will be wrapped inside a Parameter node. The order of children will be preserved.
        :param current_child: the child node to be replaced.
        :type current_child: NodeBase
        :param new_child: the new child to replace current_child with.
        :type new_child: any
        """
        if not isinstance(new_child, NodeBase):
            # wrap non-node values inside `Parameter`
            new_child = Parameter(new_child)
        if current_child not in self._children:
            raise ValueError("Cannot replace child %s because it is not a child of %s."
                                % (current_child.name, self.name))
        self._children = [_c if _c is not current_child else new_child for _c in self._children]
        new_child.add_parent(self)
        current_child.remove_parent(self)
        self.mark_for_update()

    def print_descendants(self):
        """
        Recursively print this node and its children.
        """
        NodeChildrenPrinter(self).run()


class RootNode(NodeBase):
    """
    A Node that acts as the root node of a graph. Cannot have any parent nodes. All other nodes are
    direct or indirect children of this node. This node therefore depends on all other nodes.
    """

    def __init__(self):
        self._name = '__root__'
        self._stale = True
        self._children = []
        self._parents = set()  # root node has no parents

    def add_child(self, node):
        NodeBase.add_child(self, node)

        # keep root node children sorted by name
        self._children = sorted(self._children, key=lambda x: x.name)

    def replace(self, other):
        raise TypeError("Root node cannot be replaced!")

    def add_parent(self, parent):
        raise TypeError("Cannot add parent to root node!")

    def iter_parents(self):
        # empty generator
        return iter(())

    def notify_parents(self):
        pass


@_add_common_operators
class ValueNode(NodeBase):
    """
    Partly abstract class for a graph node that can have a value.
    """

    def __init__(self, value=None, name=None):
        """
        :param value: the value of this node.
        :type value: any
        :param name: the name of this node.
        :type name: str
        """
        NodeBase.__init__(self, name=name)
        self._value = value
        self._stale = False

    def __str__(self):
        try:
            _val = str(self.value)
        except Exception as e:
            _val = repr(e)
        return "{}('{}') = {}  [{}]".format(
            self.__class__.__name__,
            self.name, _val, hex(id(self))
        )

    def _pprint(self):
        try:
            _val = str(self.value)
        except Exception as e:
            _val = repr(e)
        if '\n' in _val:
            _val = '\n' + _val
        else:
            _val = ' ' + _val
        return NodeBase._pprint(self) + ' =' + _val

    @classmethod
    def _make_binop_method(cls, op):
        """
        Define binary operators for this class.
        """
        def _op_method(self, other):
            if not isinstance(other, NodeBase):
                other = Parameter(other)
            _auto_name = "{}__{}__{}".format(self.name, op.__name__, other.name)

            return Function(
                func=op,
                name=_auto_name,
                parameters=(self, other),
            )

        return _op_method

    @classmethod
    def _make_rbinop_method(cls, op):
        """
        Define reverse binary operators for this class.
        """
        def _op_method(self, other):
            if not isinstance(other, NodeBase):
                other = Parameter(other)
            _auto_name = "{}__r{}__{}".format(self.name, op.__name__, other.name)

            return Function(
                func=op,
                name=_auto_name,
                parameters=(other, self),
            )

        return _op_method

    @classmethod
    def _make_unaryop_method(cls, op):
        """
        Define unary operators for this class.
        """
        def _op_method(self):
            _auto_name = "{}__{}".format(op.__name__, self.name)

            return Function(
                func=op,
                name=_auto_name,
                parameters=(self,),
            )

        return _op_method

    @property
    def value(self):
        """
        :return: the value of this node.
        :rtype: any
        """
        if self.stale and not self.frozen:
            self.update()

        return self._value

    @value.setter
    def value(self, value):
        self._value = value
        # notify parents that a child value has changed
        self.notify_parents()


class Empty(ValueNode):
    """
    A graph node that acts as a placeholder for a ValueNode.
    Should be replaced by a ValueNode in the finished graph.
    Raises an Exception if an attempt is made to set or retrieve the value.
    """
    def __init__(self, name=None):
        NodeBase.__init__(self, name=name)

    def __str__(self):
        return "{}('{}')  [{}]".format(
            self.__class__.__name__,
            self.name, hex(id(self))
        )

    def _pprint(self):
        return NodeBase._pprint(self)

    @property
    def value(self):
        raise TypeError(
            "Empty node '{}' does not have a value.".format(
                self.name
            )
        )

    @value.setter
    def value(self, value):
        raise TypeError(
            "Cannot set value of empty node '{}'.".format(
                self.name
            )
        )


class Parameter(ValueNode):
    """
    Simple subclass of ValueNode that represents a constant value.
    """
    def __init__(self, value, name=None):
        ValueNode.__init__(self, value=value, name=name)
        self._stale = False

    def mark_for_update(self):
        pass  # Simple parameters are always up-to-date.


class Alias(ValueNode):
    """
    A Node that only passes on the value of another node when evaluated.
    Used to add multiple functionally equivalent nodes with differing names to a graph.
    """
    def __init__(self, ref, name=None):
        ValueNode.__init__(self, value=None, name=name)

        self.ref = ref

        self._stale = True

    def _pprint(self):
        return '{} -> {}'.format(
            NodeBase._pprint(self),
            NodeBase._pprint(self.ref)
        )

    @ValueNode.value.setter
    def value(self, value):
        raise TypeError(
            "Cannot set value of alias node '{}'.".format(
                self.name
            )
        )

    @property
    def ref(self):
        """
        :return: The node for which this is an alias.
        """
        return self._children[0]

    @ref.setter
    def ref(self, ref):
        self.set_children([ref])
        ref.add_parent(self)

    def update(self):
        self._value = self.ref.value
        self._stale = False


class Function(ValueNode):
    """
    Subclass of ValueNode that computes its value from the values of its children.
    """

    def __init__(self, func, name=None, parameters=None):
        """
        Creates a new function node. If any of the parameters are not of type NodeBase they will be
        wrapped in Parameter nodes.
        :param func: the function used to compute this node's value.
        :type func: Callable
        :param name: the name of this node.
        :type name: str
        :param parameters: the parameters of the function.
        :type parameters: iterable
        """
        if name is None and func.__name__ != "<lambda>":
            name = func.__name__
        super(Function, self).__init__(value=None, name=name)

        self.func = func
        parameters = parameters or []
        self.parameters = parameters
        self.set_children(parameters)

        self._par_cache = []

    def __call__(self):
        return self._func(*self._par_cache)

    @property
    def parameters(self):
        """
        A subset of this node's children. Passed to the function handle when this node is updated.
        :return: The nodes representing the function parameters.
        """
        return self._parameters

    @parameters.setter
    def parameters(self, parameters):
        self._parameters = list(parameters)

    @property
    def func(self):
        """
        :return: The function handle to be called when updating this node.
        """
        return self._func

    @func.setter
    def func(self, function_handle):
        self._func = function_handle
        self._stale = True

    @ValueNode.value.setter
    def value(self, value):
        raise TypeError("Function node value cannot be set!")

    def add_parameter(self, parameter):
        """
        Add a parameter for this function. Automatically adds the parameter to this node's children
        as well.
        :param parameter: the parameter to be added.
        :type parameter: any
        """
        self._parameters.append(parameter)
        self.add_child(parameter)

    def update(self):
        for _child in self.get_children():
            if _child.stale and not _child.frozen:
                _child.update()
        self._par_cache = [
            _par.value
            for _par in self._parameters
        ]
        self._value = self()
        self._stale = False

    def replace(self, other, other_children=True):
        NodeBase.replace(self, other, other_children)
        if not other_children and isinstance(other, Function):
            other.parameters = self.parameters

    def replace_child(self, current_child, new_child):
        NodeBase.replace_child(self, current_child, new_child)
        self._parameters = [_p if _p is not current_child else new_child for _p in self._parameters]


class Fallback(ValueNode):
    """
    Node that tries to retrieve the value of several other nodes until it finds a node that doesn't
    raise an exception.
    """

    def __init__(self, try_nodes, exception_type=Exception, name=None):
        """
        Create a new Fallback node. Any objects in try_nodes that are not of type NodeBase will be
        wrapped inside a Parameter node.
        :param try_nodes: the value literals / nodes to try to retrieve the value from.
        :type try_nodes: iterable
        :param exception_type: the type of Exception upon which to continue to the next node.
        :type exception_type: type
        :param name: the name of the node.
        :type name: str
        """
        super(Fallback, self).__init__(value=None, name=name)
        self.try_nodes = try_nodes
        self._exception_type = exception_type
        self._value = None

        if name is None:
            self.name = '__onfail__'.join([_n.name for _n in self.try_nodes])

    def __call__(self):
        for _node in self.try_nodes:
            try:
                return _node.value
            except self._exception_type:
                pass  # function failed; try next

        raise RuntimeError(
            "Error evaluating fallback node '{}': "
            "no alternative succeeded".format(self.name)
        )

    @property
    def try_nodes(self):
        """
        :return: The nodes to try in order to retrieve the value from. Equivalent to get_children.
        :rtype: list of NodeBase
        """
        return self._children

    @try_nodes.setter
    def try_nodes(self, try_nodes):
        self.set_children(try_nodes)

    def update(self):
        self._value = self()
        self._stale = False


class Tuple(ValueNode):
    """
    Node that combines several other nodes into a tuple for its own value.
    """
    def __init__(self, nodes, name=None):
        """
        Create a new Tuple node. Objects in nodes that are not of type NodeBase will be wrapped
        inside a Parameter node.
        :param nodes: the literals / nodes to combine into a tuple.
        :type nodes: iterable
        :param name: the name of the node.
        :type name: str
        """
        super(Tuple, self).__init__(value=None, name=name)
        self.nodes = nodes
        self._value = None

        if name is None and self.nodes:
            self.name = '__'.join([_n.name for _n in self.nodes])

    def __iter__(self):
        for _n in self.nodes:
            yield _n

    def __len__(self):
        return len(self._children)

    def __getitem__(self, index):
        return self._children[index]

    def __setitem__(self, index, item):
        if not isinstance(item, NodeBase):
            item = Parameter(item)
        self._children[index] = item

    @property
    def nodes(self):
        """
        :return: The nodes representing the tuple content.
        :rtype: list of NodeBase
        """
        return self._children

    @nodes.setter
    def nodes(self, nodes):
        self.set_children(nodes)

    def update(self):
        self._value = tuple([
            _node.value for _node in self.nodes
        ])
        self._stale = False

    def iter_values(self):
        """
        Generator for Tuple.value.
        """
        for _val in self.value:
            yield _val


class Array(Tuple):
    """
    Node that combines several other nodes into a numpy array for its own value.
    """
    def __init__(self, nodes, name=None, dtype=None):
        """
        Create a new Array node. Objects in nodes that are not of type NodeBase will be wrapped
        inside a Parameter node.
        :param nodes: the literals / nodes to combine into a numpy array.
        :type nodes: iterable
        :param name: the name of the node.
        :type name: str
        :param dtype: the data type of teh numpy array.
        :type dtype: type
        """
        super(Tuple, self).__init__(value=None, name=name)
        self._dtype = dtype
        self.nodes = nodes

        if name is None and self.nodes:
            self.name = 'array_' + '__'.join([_n.name for _n in self.nodes])

    @Tuple.nodes.setter
    def nodes(self, nodes):
        self.set_children(nodes)
        self._value = np.empty(len(nodes), dtype=self._dtype)

    def update(self):
        for _i, _node in enumerate(self.nodes):
            self._value[_i] = _node.value

        self._stale = False

    def iter_values(self):
        return self.value.__iter__()


# -- Node visitors

class NodeChildrenPrinter(object):
    """
    Visitor class that recursively prints a node and its children.
    """
    def __init__(self, root_node):
        self._root = root_node

    def visit(self, node, indent=''):
        _s = node._pprint()
        if '\n' in _s:
            _s = _s.replace('\n', '\n'+indent+'    ')
        print(indent + '+ ' + _s)

    def run(self, node=None, indent=''):
        node = node if node is not None else self._root

        # visit current node
        self.visit(node, indent=indent)

        # recursively visit children
        for _c in node.get_children():
            self.run(_c, indent=indent + '  ')


class NodeSubgraphGraphvizSourceProducer(object):
    """
    Visitor class for debugging that creates a graphviz representation of a node graph
    (i.e. a Nexus). This class has **not** been unit tested.
    """
    _NODE_STYLE = {
        Function: dict(
            shape='"box"'),
        Alias: dict(
            shape='"cds"'),
        Fallback: dict(
            shape='"diamond"'),
        Empty: dict(
            shape='"Mcircle"'),
    }

    def __init__(self, root_node, exclude=None):
        self._root = root_node
        self._exclude = exclude or set()
        self._nodes = set()
        self._leaf_nodes = set()
        self._connected_nodes = set()
        self._full_edges = set()
        self._dashed_edges = set()

    def visit(self, node):
        self._nodes.add(node)
        for _c in node.get_children():
            self._nodes.add(node)
            if node in self._exclude or _c in self._exclude:
                continue
            self._full_edges.add(
                (node, _c)
            )
            self._nodes.add(_c)

        for _p in node.iter_parents():
            self._nodes.add(node)
            if node in self._exclude or _p in self._exclude:
                continue
            self._dashed_edges.add(
                (_p, node)
            )
            self._nodes.add(_p)

    def run(self, node=None, node_chain=None, out_stream=None):
        node = node if node is not None else self._root

        _first_call = node_chain is None
        node_chain = node_chain or set()
        # don't revisit previous nodes
        if node in node_chain:
            return

        # visit current node
        self.visit(node)

        node_chain.add(node)

        # recursively visit children
        for _c in node.get_children():
            NodeSubgraphGraphvizSourceProducer.run(self, _c, node_chain=node_chain)
        # recursively visit parents
        for _p in node.iter_parents():
            NodeSubgraphGraphvizSourceProducer.run(self, _p, node_chain=node_chain)

        if _first_call:

            out_stream = out_stream if out_stream is not None else sys.stdout

            import functools
            _all_edges = self._full_edges.union(self._dashed_edges)
            _connected_nodes = set(functools.reduce(
                lambda x, y: x+y,
                _all_edges,
            ))
            _isolated_nodes = self._nodes - _connected_nodes
            self._exclude |= _isolated_nodes

            _leaf_nodes = (
                set([n for _, n in _all_edges])
                -
                set([n for n, _ in _all_edges])
            )

            out_stream.write("digraph {} {{\n\n".format(node.name))
            for _n in _connected_nodes:
                if _n in self._exclude:
                    continue
                out_stream.write(
                    '  {0.name} [label=<<i>{0.__class__.__name__}</i>'
                    '<br />{0.name}>, {1}]\n'.format(
                        _n,
                        ', '.join([
                            "{}={}".format(_k, _v)
                            for _k, _v
                            in self._NODE_STYLE.get(_n.__class__, {}).items()
                        ] + (['style=filled', 'fillcolor=yellow'] if _n in _leaf_nodes else [])),
                    )
                )

            out_stream.write('\n')
            for _e in self._full_edges:
                if _e[0] in self._exclude or _e[1] in self._exclude:
                    continue
                out_stream.write("  {0.name} -> {1.name}\n".format(*_e))

            out_stream.write('\n')
            for _e in self._dashed_edges - self._full_edges:
                if _e[0] in self._exclude or _e[1] in self._exclude:
                    continue
                out_stream.write("  {0.name} -> {1.name} [style=dashed]\n".format(*_e))
            out_stream.write("\n}\n")
            out_stream.flush()

            self._full_edges = set()
            self._dashed_edges = set()


class NodeSubgraphGraphvizViewer(NodeSubgraphGraphvizSourceProducer):

    def run(self, node=None, node_chain=None):
        from graphviz import Source

        import tempfile

        out = six.StringIO()
        NodeSubgraphGraphvizSourceProducer.run(
            self, node, node_chain, out_stream=out)

        out.seek(0)
        Source(out.read()).render(tempfile.mktemp('.gv'), view=True)


class NodeCycleChecker(object):
    """
    Visitor class that ensures that a node graph is acyclic, i.e. that there are no cyclic
    dependencies.
    """
    def __init__(self, root_node):
        self._root = root_node

    def visit(self, node, seen):
        if node in seen:
            _cycle = seen[seen.index(node):] + (node,)
            raise ValueError("Dependent node cycle detected ({})".format(
                ' -> '.join([_n.name for _n in _cycle])
            ))

    def run(self, node=None, seen=None):
        node = node if node is not None else self._root
        seen = seen if seen is not None else tuple()

        # visit current node
        self.visit(node, seen)

        # keep track of nodes encountered
        seen += (node,)

        # recursively visit parents
        for _c in node.iter_parents():
            self.run(node=_c, seen=seen)


# -- Nexus

class Nexus(object):
    """
    Object representing an entire computation graph. Used in the kafe2 NexusFitter object to manage
    the caching of intermediate results for the calculation of the cost function value.
    """

    _VALID_GET_DICT_ERROR_BEHAVIORS = (
        'fail', 'none', 'exception_as_value', 'ignore', 'list')
    _VALID_ADD_EXISTING_BEHAVIORS = (
        'fail', 'ignore', 'replace', 'replace_if_alias', 'replace_if_empty')

    def __init__(self):
        self._nodes = {
            '__root__': RootNode()
        }
        self._root_ref = weakref.ref(self._nodes['__root__'])  # convenience

    def add(self, node, add_children=True, existing_behavior='fail'):
        """Add a node to the nexus.

        The `node` type should be derived from
        :py:class:`~kafe.core.fitters.nexus_v2.NodeBase`. If not, it
        will be wrapped inside a
        :py:class:`~kafe.core.fitters.nexus_v2.Parameter` node.

        Nexus node names must be unique. If a node with the same name
        has already been added, the behavior of this function can
        be configured via `existing_behavior`, which takes one of
        three values:
        - ``fail``: raise a `ValueError`
        - ``replace``: replace the existing node with the one being added
        - ``replace_if_alias``: replace the existing node with the one
            only if the existing node is an alias
        - ``replace_if_empty``: replace the existing node with the one
            only if the existing node is empty
        - ``ignore``: do not add the node

        :param node: node to add
        :type node: :py:class:`~kafe.core.fitters.nexus_v2.NodeBase` or other
        :param add_children: if True, recursively add node children to self if self does not have
        them.
        :type add_children: bool
        :param existing_behavior: how to behave if a node with the same name
            has already been added to the nexus. One of: ``fail`` (default),
            ``replace``, ``replace_if_alias``, ``replace_if_empty``,
            ``ignore``.
        :type existing_behavior: str

        :returns: the added node
        """
        # validate `existing_behavior`
        if existing_behavior not in self._VALID_ADD_EXISTING_BEHAVIORS:
            raise ValueError(
                "Unknown value {!r} for `existing_behavior`: "
                "expected one of {!r}".format(
                    existing_behavior,
                    self._VALID_ADD_EXISTING_BEHAVIORS
                )
            )

        # check if node has already been added
        if node.name in self._nodes:
            # resolve behavior if node is empty
            if existing_behavior == 'replace_if_empty':
                existing_behavior = (
                    'replace'
                    if isinstance(self._nodes[node.name], Empty)
                    else 'fail'
                )
            elif existing_behavior == 'replace_if_alias':
                existing_behavior = (
                    'replace'
                    if isinstance(self._nodes[node.name], Alias)
                    else 'fail'
                )

            # execute behavior
            if existing_behavior == 'fail':
                raise ValueError(
                    "Node '{}' already exists.".format(
                        node.name
                    )
                )
            if existing_behavior == 'replace':
                if add_children:
                    # add all dependent children to the nexus first
                    for _child in node.get_children():
                        self.add(_child, add_children=True, existing_behavior='ignore')
                # replace node
                self._nodes[node.name].replace(node)
            elif existing_behavior == 'ignore':
                return  #
            else:
                assert False  # should not get here

        # node has not been added before
        else:
            if add_children:
                # add all dependent children to the nexus first
                for _child in node.get_children():
                    self.add(_child, add_children=True, existing_behavior='ignore')

            # add new node as child of root node
            self._root_ref().add_child(node)

            # remove nexus root from all children's parents (if present)
            for _child in node.get_children():
                try:
                    self._root_ref().remove_child(_child)
                except KeyError:
                    pass

        # (re)map name to point to node
        self._nodes[node.name] = node

        # check for cycles
        NodeCycleChecker(node).run()

        return node

    def add_function(self, func, func_name=None, par_names=None, add_children=True,
                     existing_behavior='fail'):
        """Construct a `Function` node from a Python function and
        add it to the nexus.

        The object passed as `func` should be a native Python function.
        If other callables (e.g. `lambda` expressions) are passed, they
        must either have a `__name__` parameter or `func_name` must be provided.
        The function node name is set to `func_name` (if provided) or
        `func.__name__`.

        For every parameter of the function, there should already be a
        nexus node with the same name as the parameter name. If they
        exist, they will be registered automatically as dependencies of
        the function node, meaning that the function value will be
        recalculated every time a parameter changes value.

        In case the function parameter names do not correspond exactly
        to the names of the parameter nodes in the nexus, a list of
        node names to use as parameters can be specified via the
        `par_names` keyword,

        If any of the specified parameter nodes does not exist, it will
        be created. In case a default value for the parameter is
        specified in the function signature, a
        :py:class:`~kafe.core.fitters.nexus.Parameter` node will be
        created and initialized to that value. If no default is
        specified, an :py:class:`~kafe.core.fitters.nexus.Empty` node
        will be created. It should be replaced by a non-empty node
        before the function is evaluated.

        Any variable-length arguments (``*args``) and keyword arguments
        (``**kwargs``) in the function signature are ignored when
        creating parameters.

        If a node with the same name as the function node
        has already been added, the behavior of this function can
        be configured via `existing_behavior`. For details, see
        :py:meth:`~kafe.core.fitters.nexus.Nexus.add`.

        :param func: function to add as node
        :type func: function
        :param func_name: node name to use instead of function name
        :type func_name: str
        :param par_names: node name to use as function parameters
        :type par_names: list of str
        :param add_children: if True, recursively add node children to self if self does not have
        them.
        :type add_children: bool
        :param existing_behavior: how to behave if a node with the same
            name has already been added to the nexus. One of: ``fail``
            (default), ``replace``, ``replace_if_empty``, ``ignore``.
        :type existing_behavior: str

        :returns: the added function node
        """
        # resolve function name
        func_name = func_name or func.__name__

        if par_names is not None:
            _par_nodes = [
                self._nodes[_par_name] if _par_name in self._nodes else Empty(_par_name)
                for _par_name in par_names
            ]
            return self.add(
                Function(func, name=func_name, parameters=_par_nodes),
                add_children=add_children, existing_behavior=existing_behavior
            )

        # resolve parameter names
        _sig = signature(func)

        # take child node names from signature
        par_names = [_p.name for _p in _sig.parameters.values()]

        # assert compatibility between function signature
        # and supplied node names
        try:
            _ba = _sig.bind(*par_names)
            assert len(par_names) >= len(_sig.parameters)  # Assert function isn't under-supplied.
        except (AssertionError, TypeError):
            raise ValueError(
                "Error adding function: supplied "
                "parameter nodes ({}) are not compatible with "
                "function signature {!r}!".format(
                    ', '.join(map(repr, par_names)),
                    _sig
                )
            )

        _args = _ba.arguments.items()
        _args_defaults = [_a.default for _a in _sig.parameters.values()]

        assert len(_args_defaults) >= len(_args)  # barring any unbound *args, **kwargs

        # resolve parameter nodes
        _pars = []
        for (_arg_name, _node_spec), _arg_default in zip(_args, _args_defaults):

            # fail on **kwargs
            if isinstance(_node_spec, dict):
                raise ValueError(
                    "Error adding function: signature "
                    "cannot contain variable keyword arguments "
                    "(**{})!".format(
                        _arg_name
                    )
                )

            # wrap str in a tuple (normal non-* case)
            if isinstance(_node_spec, str):
                _node_spec = (_node_spec,)

            assert(isinstance(_node_spec, tuple))

            # add node or nodes under *args
            for _node_name in _node_spec:
                # check if parameter node exists
                if _node_name not in self._nodes:
                    if _arg_default == SigParameter.empty:
                        # create empty node and add
                        _pars.append(self.add(Empty(name=_node_name)))
                    else:
                        # create parameter node and add
                        _pars.append(self.add(Parameter(_arg_default, name=_node_name)))
                else:
                    _existing_node = self._nodes[_node_name]

                    # if default value provided
                    if _arg_default != SigParameter.empty:

                        # node is empty
                        if isinstance(_existing_node, Empty):
                            # value encountered for previously empty node -> replace
                            _pars.append(self.add(
                                Parameter(_arg_default, name=_existing_node.name),
                                existing_behavior='replace'
                            ))
                            continue

                        # if node non-empty
                        else:

                            # new default may conflict -> warn if necessary
                            try:
                                assert(_existing_node.value == _arg_default)
                            except AssertionError:
                                # values definitely conflict
                                warnings.warn(
                                    "Ignoring default value {!r} for function parameter '{}': "
                                    "non-empty nexus node already exists and has conflicting "
                                    "value {!r}.".format(
                                        _arg_default,
                                        _node_name,
                                        _existing_node.value
                                    ),
                                    UserWarning
                                )

                    # register existing node as parameter
                    _pars.append(_existing_node)

        # add function node to nexus and return it
        return self.add(
            Function(func, name=func_name, parameters=_pars),
            add_children=add_children, existing_behavior=existing_behavior
        )

    def add_alias(self, name, alias_for, existing_behavior='fail'):
        """Add an Alias node pointing to an existing node `alias_for`.

        If a node with the same name as the alias node
        has already been added, the behavior of this function can
        be configured via `existing_behavior`. For details, see
        :py:meth:`~kafe.core.fitters.nexus.Nexus.add`.

        :param name: name of the alias node to add
        :type name: str
        :param alias_for: name of the node being pointed to by the alias.
        :type alias_for: str
        :param existing_behavior: how to behave if a node with the same
            name has already been added to the nexus. One of: ``fail``
            (default), ``replace``, ``replace_if_empty``,
            ``replace_if_alias``, ``ignore``.
        :type existing_behavior: str

        :return: the added alias node
        """

        _alias_for_node = self.get(alias_for)
        if _alias_for_node is None:
            raise ValueError(
                "Cannot add alias for node '{}': "
                "node does not exist in nexus!".format(alias_for)
            )

        return self.add(
            Alias(
                _alias_for_node,
                name=name
            ),
            existing_behavior=existing_behavior
        )

    def add_dependency(self, name, depends_on):
        """Register a dependency between nodes explicitly.

        This indicates that the value of node `name` may depend on the
        value of another node `depends_on`. When the value of
        `depends_on` changes, `name` will be notified.

        The specification `depends_on` may be a list or a tuple, in which
        case a dependency is added for each specified node.

        Note that this method should only be used for establishing
        *explicit* dependencies. Other methods such as `add`,
        `add_function` or `add_alias` will create *implicit*
        dependencies between nodes, such as between a `Function` node
        and its parameters or between an `Alias` node and the node it
        points to). In those cases it is not necessary to call
        `add_dependency` explicitly.

        This method is useful when node values may depend on an
        external state that operates on the nexus but is not itself
        part of it.

        :param name: name of the node whose value depends on (an)other node(s)
        :type name: str
        :param depends_on: name(s) of the node(s) whose value node `name` depends on
        :type depends_on: str or tuple/list of str

        :return: self (the Nexus instance)
        :rtype: Nexus
        """
        _node = self.get(name)

        if _node is None:
            raise ValueError(
                "Cannot add dependency: dependent node '{}' does "
                "not exist!".format(name)
            )

        if not isinstance(depends_on, (tuple, list)):
            depends_on = (depends_on,)

        _not_found = [_dep for _dep in depends_on if self.get(_dep) is None]

        if _not_found:
            raise ValueError(
                "Cannot add dependency: the following nodes passed to "
                "`depends_on` do not exist: {}".format(
                    ', '.join(map(repr, _not_found))
                )
            )

        # add dependent node `name` as a parent of each node in `depends_on`
        for _dep in depends_on:
            _node.add_child(self.get(_dep))

        # check for cycles
        NodeCycleChecker(_node).run()

    def get(self, node_name):
        """Retrieve a node by its name or ``None`` if no such node exists.

        :param node_name: name of the node to get
        :type node_name: str

        :return: the node or ``None`` if it does not exist
        """
        return self._nodes.get(node_name, None)

    def get_value_dict(self, node_names=None, error_behavior='fail'):
        """Return a mapping of node names to their current values.

        The behavior when encountering an evaluation error can be
        specified as a string:
        *  ``'fail'``: raise an exception
        *  ``'none'``: use ``None`` as a value
        *  ``'exception_as_value'``: use the raised exception as a node value
        *  ``'ignore'``: do not include erroring nodes in result dict
        *  ``'list'``: return list of erroring nodes in result dict
            under special key ``__error__``

        :param node_names: name of the nodes for which to get values.
            If ``None``, all nodes are evaluated.
        :type node_names: list of str or ``None``
        :param error_behavior: how to behave if a node with the same name
            has already been added to the nexus. One of: ``fail`` (default),
            ``none``, ``ignore``, ``list``, ``exception_as_value``.
        :type error_behavior: str

        :return: dict mapping node names to values
        """

        # validate `error_behavior`
        if error_behavior not in self._VALID_GET_DICT_ERROR_BEHAVIORS:
            raise ValueError(
                "Unknown value {!r} for `error_behavior`: "
                "expected one of {!r}".format(
                    error_behavior,
                    self._VALID_GET_DICT_ERROR_BEHAVIORS
                )
            )

        # construct result dict
        _result_dict = {}
        for _name, _node in self._nodes.items():
            # skip root node (has no value anyway)
            if _name == '__root__':
                continue
            if node_names is not None and _name not in node_names:
                continue

            # attempt evaluation and behave accordingly on error
            try:
                _val = _node.value
            except Exception as e:
                if error_behavior == 'fail':
                    raise e
                if error_behavior == 'none':
                    _val = None
                elif error_behavior == 'exception_as_value':
                    _val = e
                elif error_behavior == 'ignore':
                    continue
                elif error_behavior == 'list':
                    _result_dict.setdefault('__error__', []).append(_name)
                    continue
                else:
                    assert False, "Something went terribly wrong. " \
                                  "Unknown error behaviour {}".format(error_behavior)
            else:
                _result_dict[_name] = _val

        return _result_dict

    def print_state(self):
        """Print a representation of the nexus state."""
        NodeChildrenPrinter(self._root_ref()).run()



"""This submodule provides the necessary objects for parameter estimation using data consisting of ordered *xy* pairs.
This fit type is used for most cases e.g. when performing fits for the first time or in physics laboratory courses.

:synopsis: This submodule provides the necessary objects for parameter estimation using data consisting of ordered *xy*
    pairs.

.. moduleauthor:: Daniel Savoiu <daniel.savoiu@cern.ch>
"""

from .container import *
from .cost import *
from .ensemble import *
from .fit import *
from .model import *
from .plot import *



from copy import deepcopy

from .._base import FitBase, DataContainerBase
from .container import HistContainer
from .._base.cost import CostFunction_NegLogLikelihood
from .model import HistParametricModel, HistModelFunction
from .plot import HistPlotAdapter
from ..util import function_library, collect

__all__ = ['HistFit']


class HistFit(FitBase):
    CONTAINER_TYPE = HistContainer
    MODEL_TYPE = HistParametricModel
    MODEL_FUNCTION_TYPE = HistModelFunction
    PLOT_ADAPTER_TYPE = HistPlotAdapter
    RESERVED_NODE_NAMES = {'data', 'model', 'model_density', 'cost',
                          'data_error', 'model_error', 'total_error',
                          'data_cov_mat', 'model_cov_mat', 'total_cov_mat',
                          'data_cor_mat', 'model_cor_mat', 'total_cor_mat'}
    _BASIC_ERROR_NAMES = {'data_error', 'model_error', 'data_cov_mat', 'model_cov_mat'}

    def __init__(self,
                 data,
                 model_function=function_library.normal_distribution,
                 cost_function=CostFunction_NegLogLikelihood(
                    data_point_distribution='poisson'),
                 bin_evaluation="simpson",
                 density=True,
                 minimizer=None,
                 minimizer_kwargs=None,
                 dynamic_error_algorithm="nonlinear"):
        """
        Construct a fit of a model to a histogram. If bin_evaluation is a Python function or
        of a numpy.vectorize object it is interpreted as the antiderivative of
        model_density_function. If bin_evaluation is equal to "rectangle", "midpoint", "trapezoid",
        or "simpson" the bin heights are evaluated according to the corresponding quadrature
        formula. If bin_evaluation is equal to "numerical" the bin heights are evaluated by
        numerically integrating model_density_function.

        :param data: a :py:class:`~kafe2.fit.hist.HistContainer` representing histogrammed data
        :type data: :py:class:`~kafe2.fit.hist.HistContainer`
        :param model_function: the model (density) function
        :type model__function: :py:class:`~kafe2.fit.hist.HistModelFunction` or unwrapped
            native Python function
        :param cost_function: the cost function
        :type cost_function: :py:class:`~kafe2.fit._base.CostFunctionBase`-derived or unwrapped
            native Python function
        :param bin_evaluation: how the model evaluates bin heights.
        :type bin_evaluation: str, callable, or numpy.vectorize
        :param density: if True, scale model function to the number of data points.
        :type density: bool
        :param minimizer: the minimizer to use for fitting.
        :type minimizer: None, "iminuit", "tminuit", or "scipy".
        :param minimizer_kwargs: dictionary with kwargs for the minimizer.
        :type minimizer_kwargs: dict
        """
        self._bin_evaluation = bin_evaluation
        self._density = density
        super(HistFit, self).__init__(
            data=data, model_function=model_function, cost_function=cost_function,
            minimizer=minimizer, minimizer_kwargs=minimizer_kwargs,
            dynamic_error_algorithm=dynamic_error_algorithm)

    # -- private methods

    def _init_nexus(self):
        super(HistFit, self)._init_nexus()

        self._nexus.add_dependency(
            'model',
            depends_on=(
                'parameter_values'
            )
        )

    def _set_new_data(self, new_data):
        if isinstance(new_data, self.CONTAINER_TYPE):
            self._data_container = deepcopy(new_data)
        elif isinstance(new_data, DataContainerBase):
            raise TypeError("Incompatible container type '{}' (expected '{}')"
                            .format(type(new_data), self.CONTAINER_TYPE))
        else:
            raise TypeError("Fitting a histogram requires a HistContainer!")
        self._data_container._on_error_change_callback = self._on_error_change

        self._nexus.get('data').mark_for_update()

    def _set_new_parametric_model(self):
        # create the child ParametricModel object
        self._param_model = HistParametricModel(
            self._data_container.size, self._data_container.bin_range, self._model_function,
            self.parameter_values, self._data_container.bin_edges,
            bin_evaluation=self._bin_evaluation, density=self._density)

    # -- public properties

    @property
    def model(self):
        """array of model predictions for the data points"""
        self._param_model.parameters = self.parameter_values  # this is lazy, so just do it
        # if model is just a density, scale it to the number of data points
        if self._param_model.density:
            return self._param_model.data * self._data_container.n_entries
        else:
            return self._param_model.data

    @property
    def density(self):
        return self._param_model.density

    # FIXME: how to handle scaling of model_error

    # -- public methods

    ## add_error... methods inherited from FitBase ##

    def eval_model_function_density(self, x, model_parameters=None):
        """
        Evaluate the model function density.

        :param x: values of *x* at which to evaluate the model function density
        :type x: iterable of float
        :param model_parameters: the model parameter values (if ``None``, the current values are used)
        :type model_parameters: iterable of float
        :return: model function density values
        :rtype: :py:class:`numpy.ndarray`
        """
        self._param_model.parameters = self.parameter_values  # this is lazy, so just do it
        return self._param_model.eval_model_function_density(x=x, model_parameters=model_parameters)



from __future__ import print_function
import six
import ctypes
from .minimizer_base import MinimizerBase
from ..contour import ContourFactory
try:
    from ROOT import TMinuit, Double, Long
    from ROOT import TMath  # for using ROOT's chi2prob function
except ImportError:
    # TODO: handle importing nonexistent minimizer
    raise

from array import array as arr  # array needed for TMinuit arguments

import numpy as np


class MinimizerROOTTMinuit(MinimizerBase):
    def __init__(self,
                 parameter_names, parameter_values, parameter_errors,
                 function_to_minimize, tolerance=1e-9, errordef=MinimizerBase.ERRORDEF_CHI2,
                 strategy=1):
        self._strategy = strategy

        self._par_bounds = np.array([None] * len(parameter_names))
        self._par_fixed = np.array([False] * len(parameter_names))

        self.reset()  # sets self.__gMinuit and caches to None
        super(MinimizerROOTTMinuit, self).__init__(
            parameter_names=parameter_names, parameter_values=parameter_values,
            parameter_errors=parameter_errors, function_to_minimize=function_to_minimize,
            tolerance=tolerance, errordef=errordef
        )

    # -- private methods

    def _save_state(self):
        if self._par_val is None:
            self._save_state_dict["par_val"] = self._par_val
        else:
            self._save_state_dict["par_val"] = np.array(self._par_val)
        if self._par_err is None:
            self._save_state_dict["par_err"] = self._par_err
        else:
            self._save_state_dict["par_err"] = np.array(self._par_err)
        self._save_state_dict['par_fixed'] = np.array(self._par_fixed)
        self._save_state_dict['gMinuit'] = self.__gMinuit
        super(MinimizerROOTTMinuit, self)._save_state()

    def _load_state(self):
        self.reset()
        self._par_val = self._save_state_dict["par_val"]
        if self._par_val is not None:
            self._par_val = np.array(self._par_val)
        self._par_err = self._save_state_dict["par_err"]
        if self._par_err is not None:
            self._par_err = np.array(self._par_err)
        self._par_fixed = np.array(self._save_state_dict['par_fixed'])
        self.__gMinuit = self._save_state_dict['gMinuit']
        # call the function to propagate the changes to the nexus:
        self._func_handle(*self.parameter_values)
        super(MinimizerROOTTMinuit, self)._load_state()

    def _recreate_gMinuit(self):
        self.__gMinuit = TMinuit(self.num_pars)
        self.__gMinuit.SetPrintLevel(-1)
        self.__gMinuit.mncomd("SET STRATEGY {}".format(self._strategy), ctypes.c_int(0))
        self.__gMinuit.SetFCN(self._minuit_fcn)
        self.__gMinuit.SetErrorDef(self._err_def)

        # set gMinuit parameters
        error_code = ctypes.c_int(0)
        for _pid, (_pn, _pv, _pe) in enumerate(zip(self._par_names, self._par_val, self._par_err)):
            self.__gMinuit.mnparm(_pid,
                                  _pn,
                                  _pv,
                                  0.1 * _pe,
                                  0, 0, error_code)

        err_code = ctypes.c_int(0)
        # set fixed parameters
        for _par_id, _pf in enumerate(self._par_fixed):
            if _pf:
                self.__gMinuit.mnfixp(_par_id, err_code)

        # set parameter limits
        for _par_id, _pb in enumerate(self._par_bounds):
            if _pb is not None:
                _lo_lim, _up_lim = _pb
                self.__gMinuit.mnexcm("SET LIM",
                                      arr('d', [_par_id + 1, _lo_lim, _up_lim]), 3, error_code)

    def _get_gMinuit(self):
        if self.__gMinuit is None:
            self._recreate_gMinuit()
        return self.__gMinuit

    def _migrad(self, max_calls=6000):
        # need to set the FCN explicitly before every call
        self._get_gMinuit().SetFCN(self._minuit_fcn)
        error_code = ctypes.c_int(0)
        self._get_gMinuit().mnexcm("MIGRAD",
                                    arr('d', [max_calls, self.tolerance]),
                                    2, error_code)

    def _minuit_fcn(self,
                    number_of_parameters, derivatives, f, parameters, internal_flag):
        """
        This is actually a function called in *ROOT* and acting as a C wrapper
        for our `FCN`, which is implemented in Python.

        This function is called by `Minuit` several times during a fitters. It
        doesn't return anything but modifies one of its arguments (*f*).
        This is *ugly*, but it's how *ROOT*'s ``TMinuit`` works. Its argument
        structure is fixed and determined by `Minuit`:

        **number_of_parameters** : int
            The number of parameters of the current fitters

        **derivatives** : C array
            If the user chooses to calculate the first derivative of the
            function inside the `FCN`, this value should be written here. This
            interface to `Minuit` ignores this derivative, however, so
            calculating this inside the `FCN` has no effect (yet).

        **f** : C array
            The desired function value is in f[0] after execution.

        **parameters** : C array
            A C array of parameters. Is cast to a Python list

        **internal_flag** : int
            A flag allowing for different behaviour of the function.
            Can be any integer from 1 (initial run) to 4(normal run). See
            `Minuit`'s specification.
        """

        # Retrieve the parameters from the C side of ROOT and
        # store them in a Python list -- resource-intensive
        # for many calls, but can't be improved (yet?)
        parameter_list = np.frombuffer(parameters,
                                       dtype=float,
                                       count=self.num_pars)

        # call the Python implementation of FCN.
        f[0] = self._func_wrapper(*parameter_list)

    def _calculate_asymmetric_parameter_errors(self):
        self._get_gMinuit().mnmnos()
        _asymm_par_errs = np.zeros(shape=(self.num_pars, 2))
        for _n in range(self.num_pars):
            _number = Long(_n)
            _eplus = ctypes.c_double(0)
            _eminus = ctypes.c_double(0)
            _eparab = ctypes.c_double(0)
            _gcc = ctypes.c_double(0)
            self._get_gMinuit().mnerrs(_number, _eplus, _eminus, _eparab, _gcc)
            _asymm_par_errs[_n, 0] = _eminus.value
            _asymm_par_errs[_n, 1] = _eplus.value
        self.minimize()
        return _asymm_par_errs

    def _get_fit_info(self, info):
        '''Retrieves other info from `Minuit`.
        **info** : string
            Information about the fit to retrieve.
            This can be any of the following:
              - ``'fcn'``: `FCN` value at minimum,
              - ``'edm'``: estimated distance to minimum
              - ``'err_def'``: `Minuit` error matrix status code
              - ``'status_code'``: `Minuit` general status code
        '''

        # declare vars in which to retrieve other info
        fcn_at_min = ctypes.c_double(0)
        edm = ctypes.c_double(0)
        err_def = ctypes.c_double(0)
        n_var_param = ctypes.c_int(0)
        n_tot_param = ctypes.c_int(0)
        status_code = ctypes.c_int(0)

        # Tell TMinuit to update the variables declared above
        self.__gMinuit.mnstat(fcn_at_min, edm, err_def, n_var_param, n_tot_param, status_code)

        if info == 'fcn':
            return fcn_at_min.value
        elif info == 'edm':
            return edm.value
        elif info == 'err_def':
            return err_def.value
        elif info == 'status_code':
            return status_code.value
        else:
            raise ValueError("Unknown fit info: %s" % info)

    # -- public properties

    @property
    def hessian(self):
        if not self.did_fit:
            return None
        if self._hessian is None:
            _submat = self._remove_zeroes_for_fixed(self.cov_mat)
            _submat_inv = 2.0 * self.errordef * np.linalg.inv(_submat)
            self._hessian = self._fill_in_zeroes_for_fixed(_submat_inv)
        return self._hessian.copy()

    @property
    def cov_mat(self):
        if not self.did_fit:
            return None
        if self._par_cov_mat is None:
            _n_pars_total = self.num_pars
            _tmp_mat_array = arr('d', [0.0]*(_n_pars_total**2))
            # get parameter covariance matrix from TMinuit
            self._get_gMinuit().mnemat(_tmp_mat_array, _n_pars_total)
            # reshape into 2D array
            _sub_cov_mat = np.asarray(
                np.reshape(
                    _tmp_mat_array,
                    (_n_pars_total, _n_pars_total)
                ),
                dtype=np.float
            )
            _num_pars_free = np.sum(np.invert(self._par_fixed))
            _sub_cov_mat = _sub_cov_mat[:_num_pars_free, :_num_pars_free]
            self._par_cov_mat = self._fill_in_zeroes_for_fixed(_sub_cov_mat)
        return self._par_cov_mat.copy()

    @property
    def hessian_inv(self):
        if not self.did_fit:
            return None
        if self._hessian_inv is None:
            self._hessian_inv = self.cov_mat / (2.0 * self.errordef)
        return self._hessian_inv.copy()

    @property
    def parameter_values(self):
        return self._par_val.copy()

    @parameter_values.setter
    def parameter_values(self, new_values):
        self._par_val = np.array(new_values)
        self.reset()

    @property
    def parameter_errors(self):
        return self._par_err.copy()

    @parameter_errors.setter
    def parameter_errors(self, new_errors):
        _err_array = np.array(new_errors)
        if not np.all(_err_array > 0):
            raise ValueError("All parameter errors must be > 0! Received: %s" % new_errors)
        self._par_err = _err_array
        self.reset()

    # -- private "properties"

    # -- public methods

    def reset(self):
        super(MinimizerROOTTMinuit, self).reset()
        self.__gMinuit = None

    def set(self, parameter_name, parameter_value):
        if parameter_name not in self._par_names:
            raise ValueError("No parameter named '%s'!" % (parameter_name,))
        _par_id = self.parameter_names.index(parameter_name)
        self._par_val[_par_id] = parameter_value
        self.reset()

    def fix(self, parameter_name):
        # set local flag
        _par_id = self.parameter_names.index(parameter_name)
        if self._par_fixed[_par_id]:
            return  # par is already fixed
        self._par_fixed[_par_id] = True
        if self.__gMinuit is not None:
            # also update Minuit instance
            err_code = ctypes.c_int(0)
            self.__gMinuit.mnfixp(_par_id, err_code)
            # self.__gMinuit.mnexcm("FIX",
            #                   arr('d', [_par_id+1]), 1, error_code)
        self._invalidate_cache()

    def is_fixed(self, parameter_name):
        _par_id = self.parameter_names.index(parameter_name)
        return self._par_fixed[_par_id]

    def release(self, parameter_name):
        # set local flag
        _par_id = self.parameter_names.index(parameter_name)
        if not self._par_fixed[_par_id]:
            return  # par is already released
        self._par_fixed[_par_id] = False
        if self.__gMinuit is not None:
            # also update Minuit instance
            self.__gMinuit.mnfree(-_par_id-1)
            # self.__gMinuit.mnexcm("RELEASE",
            #                   arr('d', [_par_id+1]), 1, error_code)
        self._invalidate_cache()

    def limit(self, parameter_name, parameter_bounds):
        assert len(parameter_bounds) == 2
        if parameter_bounds[0] is None or parameter_bounds[1] is None:
            raise RuntimeError(
                "Cannot define one-sided parameter limits when using the ROOT TMinuit Minimizer.")
        # set local flag
        _par_id = self.parameter_names.index(parameter_name)
        if self._par_bounds[_par_id] == parameter_bounds:
            return  # same limits already set
        if self._par_val[_par_id] < parameter_bounds[0]:
            self.set(parameter_name, parameter_bounds[0])
        elif self._par_val[_par_id] > parameter_bounds[1]:
            self.set(parameter_name, parameter_bounds[1])
        self._par_bounds[_par_id] = parameter_bounds
        if self.__gMinuit is not None:
            _lo_lim, _up_lim = self._par_bounds[_par_id]
            # also update Minuit instance
            error_code = ctypes.c_int(0)
            self.__gMinuit.mnexcm("SET LIM",
                     arr('d', [_par_id+1, _lo_lim, _up_lim]), 3, error_code)
            self._did_fit = False
        self._invalidate_cache()

    def unlimit(self, parameter_name):
        # set local flag
        _par_id = self.parameter_names.index(parameter_name)
        if self._par_bounds[_par_id] is None:
            return  # parameter is already unlimited
        self._par_bounds[_par_id] = None
        if self.__gMinuit is not None:
            # also update Minuit instance
            error_code = ctypes.c_int(0)
            self.__gMinuit.mnexcm("SET LIM",
                     arr('d', [_par_id+1]), 1, error_code)
            self._did_fit = False
        self._invalidate_cache()

    def minimize(self, max_calls=6000):
        if np.all(self._par_fixed):
            raise RuntimeError("Cannot perform a fit if all parameters are fixed!")
        self._migrad(max_calls=max_calls)

        # retrieve fitters parameters
        self._par_val = np.zeros(self.num_pars)
        self._par_err = np.zeros(self.num_pars)
        _pv, _pe = ctypes.c_double(0), ctypes.c_double(0)
        for _par_id in six.moves.range(0, self.num_pars):
            self.__gMinuit.GetParameter(_par_id, _pv, _pe)  # retrieve fit result
            self._par_val[_par_id] = _pv.value
            self._par_err[_par_id] = _pe.value

        self._did_fit = True
        
    def contour(self, parameter_name_1, parameter_name_2, sigma=1.0, **minimizer_contour_kwargs):
        if not self.did_fit:
            raise RuntimeError("Need to perform a fit before calling contour()!")
        _numpoints = minimizer_contour_kwargs.pop("numpoints", 100)
        if minimizer_contour_kwargs:
            raise ValueError("Unknown parameters: {}".format(minimizer_contour_kwargs))
        _id_1 = self.parameter_names.index(parameter_name_1)
        _id_2 = self.parameter_names.index(parameter_name_2)
        self.__gMinuit.SetErrorDef(sigma ** 2)
        _t_graph = self.__gMinuit.Contour(_numpoints, _id_1, _id_2)
        self.__gMinuit.SetErrorDef(self._err_def)
        
        _x_buffer, _y_buffer = _t_graph.GetX(), _t_graph.GetY()
        _N = _t_graph.GetN()
        
        _x = np.frombuffer(_x_buffer, dtype=float, count=_N)
        _y = np.frombuffer(_y_buffer, dtype=float, count=_N)
        self._func_handle(*self.parameter_values)
        return ContourFactory.create_xy_contour((_x, _y), sigma)
    
    def profile(self, parameter_name, bins=21, bound=2, args=None, subtract_min=False):
        if not self.did_fit:
            raise RuntimeError("Need to perform a fit before calling profile()!")
        
        MAX_ITERATIONS = 6000
        
        _error_code = ctypes.c_int(0)
        _minuit_id = Long(self.parameter_names.index(parameter_name) + 1)

        _par_min = ctypes.c_double(0)
        _par_err = ctypes.c_double(0)
        self.__gMinuit.GetParameter(_minuit_id - 1, _par_min, _par_err)
        _par_min = _par_min.value
        _par_err = _par_err.value

        _x = np.linspace(start=_par_min - bound * _par_err, stop=_par_min + bound * _par_err, num=bins, endpoint=True)

        self.__gMinuit.mnexcm("FIX", arr('d', [_minuit_id]), 1, _error_code)

        _y = np.zeros(bins)
        for i in range(bins):
            self.__gMinuit.mnexcm("SET PAR", arr('d', [_minuit_id, Double(_x[i])]), 2, _error_code)
            self.__gMinuit.mnexcm("MIGRAD", arr('d', [MAX_ITERATIONS, self.tolerance]), 2, _error_code)
            _y[i] = self._get_fit_info("fcn")

        self.__gMinuit.mnexcm("RELEASE", arr('d', [_minuit_id]), 1, _error_code)
        self._migrad()
        self.__gMinuit.mnexcm("SET PAR", arr('d', [_minuit_id, Double(_par_min)]), 2, _error_code)

        if subtract_min:
            _y -= self.function_value

        return np.asarray((_x, _y))



import unittest
import numpy as np
import scipy.stats

from kafe2.fit.tools.ensemble import (broadcast_to_shape,
                                     EnsembleVariable, EnsembleVariableProbabilityDistribution,
                                     EnsembleError)


class TestCustomBroadcast(unittest.TestCase):

    def setUp(self):
        pass

    def test_compare_shape_scheme_default(self):
        # default numpy broadcasting scheme:
        #        shape:  a x b x c
        #        array:      b x 1
        #      ---------------------
        #       result:  a x b x c
        _array_shape = (4, 6)
        _array = np.empty(_array_shape)
        _target_shape = (5, 3, 4, 6)
        _result_shape = _target_shape
        _broadcasted_array = broadcast_to_shape(_array, _target_shape, scheme='default')
        self.assertEqual(_broadcasted_array.shape, _result_shape)

    def test_compare_shape_scheme_transposed(self):
        # "transposed" numpy broadcasting scheme:
        # (broadcasting is done on the transposed arrays
        #  and the result is transposed back)
        #        shape:  a x b x c
        #        array:  a x 1
        #      ---------------------
        #       result:  a x b x c
        _array_shape = (5, 3)
        _array = np.empty(_array_shape)
        _target_shape = (5, 3, 4, 6)
        _result_shape = _target_shape
        _broadcasted_array = broadcast_to_shape(_array, _target_shape, scheme='transposed')
        self.assertEqual(_broadcasted_array.shape, _result_shape)

    def test_compare_shape_scheme_expand_left(self):
        # `len(shape)` axes are added to the array shape (to the left)
        # and then the 'transposed' broadcasting is applied
        #        shape:  a x b x c          (self.ndim = 3)
        #        array:  1 x 1 x 1 x n x m
        #      -----------------------------
        #       result:  a x b x c x n x m
        _array_shape = (2, 9)
        _array = np.empty(_array_shape)
        _target_shape = (5, 3, 4, 6)
        _result_shape = (5, 3, 4, 6, 2, 9)
        _broadcasted_array = broadcast_to_shape(_array, _target_shape, scheme='expand_left')
        self.assertEqual(_broadcasted_array.shape, _result_shape)

    def test_compare_shape_scheme_expand_right(self):
        # `len(shape)` axes are added to the array shape (to the right)
        # and then broadcasting is done normally
        #        shape:          a x b x c  (self.ndim = 3)
        #        array:  n x m x 1 x 1 x 1
        #      -----------------------------
        #       result:  n x m x a x b x c
        _array_shape = (2, 9)
        _array = np.empty(_array_shape)
        _target_shape = (5, 3, 4, 6)
        _result_shape = (2, 9, 5, 3, 4, 6)
        _broadcasted_array = broadcast_to_shape(_array, _target_shape, scheme='expand_right')
        self.assertEqual(_broadcasted_array.shape, _result_shape)

    def test_compare_shape_scheme_expand_left_successive(self):
        # dimensions are added to the array (on the left) until
        # a broadcastable situation is reached
        # target shape to the array shape:
        #        shape:  a x b
        #        array:      b x n
        #      ---------------------
        #       result:  1 x b x n
        _array_shape = (2, 9, 3)
        _array = np.empty(_array_shape)
        _target_shape = (4, 6, 2, 9)
        _result_shape = (1, 1, 2, 9, 3)
        _broadcasted_array = broadcast_to_shape(_array, _target_shape, scheme='expand_left_successive')
        self.assertEqual(_broadcasted_array.shape, _result_shape)

    def test_compare_shape_scheme_expand_right_successive(self):
        # dimensions are added to the array (on the right) until
        # a broadcastable situation is reached
        # target shape to the array shape:
        #        shape:      a x b
        #        array:  n x a
        #      ---------------------
        #       result:  n x a x 1
        _array_shape = (7, 4, 6)
        _array = np.empty(_array_shape)
        _target_shape = (4, 6, 2, 9)
        _result_shape = (7, 4, 6, 1, 1)
        _broadcasted_array = broadcast_to_shape(_array, _target_shape, scheme='expand_right_successive')
        self.assertEqual(_broadcasted_array.shape, _result_shape)


@unittest.skip("Problem with collections.Sequence")
class TestEnsembleVariable(unittest.TestCase):

    def setUp(self):
        self._ref_size = 1000
        self._ref_shape = (4,7)

        self._ref_total_shape = tuple([self._ref_size] + list(self._ref_shape))

        self._ref_loc = 12.
        self._ref_scale = 3.

        np.random.seed(123456)

        _total_entries = np.prod(self._ref_total_shape)
        self._ref_array = np.random.normal(loc=self._ref_loc, scale=self._ref_scale, size=_total_entries)
        self._ref_array = self._ref_array.reshape(self._ref_total_shape)

        self.ev_no_dist = EnsembleVariable(ensemble_array=self._ref_array,
                                           distribution=None, distribution_parameters=None)
        self.ev_with_dist = EnsembleVariable(ensemble_array=self._ref_array,
                                             distribution=scipy.stats.norm,
                                             distribution_parameters=dict(loc=self._ref_loc, scale=self._ref_scale))

        self._ref_observed_means = np.mean(self._ref_array, axis=0)
        self._ref_observed_stds = np.std(self._ref_array, axis=0)

        self._ref_pdf_frozen = scipy.stats.norm(loc=self._ref_loc, scale=self._ref_scale)
        self._ref_pdf_eval_x = np.linspace(-3, 3, 20)
        self._ref_pdf_eval_y = self._ref_pdf_frozen.pdf(self._ref_pdf_eval_x)
        self._ref_pdf_eval_x_in_shape = np.zeros(self._ref_shape)
        self._ref_pdf_eval_y_in_shape = self._ref_pdf_frozen.pdf(0) * np.ones(self._ref_shape)

    def test_compare_size(self):
        self.assertEqual(self.ev_no_dist.size, self._ref_size)
        self.assertEqual(self.ev_with_dist.size, self._ref_size)

    def test_compare_shape(self):
        self.assertEqual(self.ev_no_dist.shape, self._ref_shape)
        self.assertEqual(self.ev_with_dist.shape, self._ref_shape)

    def test_compare_observed_means(self):
        self.assertTrue(
            np.allclose(
                self.ev_no_dist.mean,
                self._ref_observed_means
            )
        )
        self.assertTrue(
            np.allclose(
                self.ev_with_dist.mean,
                self._ref_observed_means
            )
        )

    def test_compare_observed_stds(self):
        self.assertTrue(
            np.allclose(
                self.ev_no_dist.std,
                self._ref_observed_stds
            )
        )
        self.assertTrue(
            np.allclose(
                self.ev_with_dist.std,
                self._ref_observed_stds
            )
        )

    def test_compare_expected_means(self):
        self.assertTrue(
            np.allclose(
                self.ev_with_dist.dist.mean,
                self._ref_loc
            )
        )

    def test_compare_expected_stds(self):
        self.assertTrue(
            np.allclose(
                self.ev_with_dist.dist.std,
                self._ref_scale
            )
        )

    def test_compare_dist(self):
        self.assertIs(self.ev_no_dist.dist, None)
        self.assertTrue(
            isinstance(self.ev_with_dist.dist, EnsembleVariableProbabilityDistribution)
        )

    def test_compare_dist_eval_pdf_scalar_x(self):
        _eval_y_compare = self.ev_with_dist.dist.eval(0, x_contains_var_shape=False)
        self.assertTrue(
            np.allclose(
                _eval_y_compare,
                self._ref_pdf_frozen.pdf(0)
            )
        )

    def test_compare_dist_eval_pdf_vector_x(self):
        _eval_y_compare = self.ev_with_dist.dist.eval(self._ref_pdf_eval_x, x_contains_var_shape=False)
        self.assertTrue(
            np.allclose(
                self._ref_pdf_eval_y,
                _eval_y_compare
            )
        )

    def test_compare_dist_eval_pdf_in_shape(self):
        _eval_y_compare = self.ev_with_dist.dist.eval(self._ref_pdf_eval_x_in_shape, x_contains_var_shape=True)
        self.assertTrue(
            np.allclose(
                self._ref_pdf_eval_y_in_shape,
                _eval_y_compare
            )
        )



import numpy as np

from .._base import CostFunction, CostFunction_Chi2

__all__ = ['SharedCostFunction', 'MultiCostFunction']


class SharedCostFunction(CostFunction_Chi2):
    def __init__(self, fallback_on_singular=True):
        self._DATA_NAME = "y_data"
        self._MODEL_NAME = "y_model"
        self._COV_MAT_CHOLESKY_NAME = "total_cov_mat_cholesky"
        super(SharedCostFunction, self).__init__(
            errors_to_use="covariance", fallback_on_singular=fallback_on_singular,
            add_constraint_cost=False, add_determinant_cost=True)


class MultiCostFunction(CostFunction):

    def __init__(self, singular_cost_functions, cost_function_names):
        super(MultiCostFunction, self).__init__(
            cost_function=MultiCostFunction.cost_sum,
            arg_names=cost_function_names,
            add_constraint_cost=True,
            add_determinant_cost=False
        )
        self._needs_errors = np.any([_scf.needs_errors for _scf in singular_cost_functions])
        self._saturated = np.all([_scf.saturated for _scf in singular_cost_functions])
        self._is_chi2 = np.all([_scf.is_chi2 for _scf in singular_cost_functions])
        if self.is_chi2:
            self.formatter.name = "chi2"
            self.formatter.latex_name = "\\chi^2"
        else:
            self.formatter.name = "cost"
            self.formatter.latex_name = None

    @staticmethod
    def cost_sum(*single_costs):
        return np.sum(single_costs)



import abc
import six
from .nexus import Nexus
from .nexus_fitter import NexusFitter
from ...config import kc

__all__ = ['get_fitter', 'Nexus', 'NexusFitter']

AVAILABLE_FITTERS = {'nexus_fitter': NexusFitter}


# try:
#     from .simple_fitter import SimpleFitter, SimpleFitterException
#     __all__ += ['SimpleFitter', 'SimpleFitterException']
#
#     AVAILABLE_FITTERS.update({
#         'simple': SimpleFitter,
#     })
# except ImportError:
#     pass


def get_fitter(fitter_spec):
    global AVAILABLE_FITTERS

    # for 'None', return the default fitter
    if fitter_spec is None:
        _fitter = AVAILABLE_FITTERS.get(kc['core']['fitters']['default_fitter'], None)
    else:
        fitter_spec = fitter_spec.lower()
        _fitter = AVAILABLE_FITTERS.get(fitter_spec, None)

    if _fitter is None:
        raise ValueError("Unknown fitter '{}'! Available: {}".format(fitter_spec, AVAILABLE_FITTERS.keys()))

    return _fitter


@six.add_metaclass(abc.ABCMeta)
class FitterBase(object):
    """
    Purely abstract class. Defines the minimal interface required by all specializations.
    """

    @abc.abstractmethod
    def do_fit(self):
        pass

    @abc.abstractmethod
    def _fcn_wrapper(self):
        pass



import abc

from ..io import InputFileHandle, OutputFileHandle


class FileIOMixin(object):
    """
    Mixin class for kafe2 objects that allows them to be read from/written to files
    using the *kafe2.fit.representation* module.
    """

    def __init__(self, *args, **kwargs):
        super(FileIOMixin, self).__init__(*args, **kwargs)

    @abc.abstractmethod
    def _get_base_class(self):
        pass

    @abc.abstractmethod
    def _get_object_type_name(self):
        pass

    @staticmethod
    def _get_file_format(filename, file_format=None):
        _basename_ext = filename.split('.')
        if len(_basename_ext) > 1:
            _basename, _ext = _basename_ext[:-1], _basename_ext[-1]
        else:
            _basename, _ext = _basename_ext[0], None

        if file_format is None and _ext is None:
            raise ValueError("Cannot detect file format from "
                                  "filename '{}' and no format specified!".format(filename))
        return file_format or _ext  # choose 'format' if specified, otherwise use filename extension

    @classmethod
    def from_file(cls, filename, file_format=None):
        """Read kafe2 object from file"""
        from ..representation import get_reader

        _format = cls._get_file_format(filename=filename, file_format=file_format)
        _reader_class = get_reader(cls._get_object_type_name(), _format)
        _object = _reader_class(InputFileHandle(filename=filename)).read()

        # check if the container is the right type (do not check if calling from DataContainerBase)
        if not _object.__class__ == cls and not cls == cls._get_base_class():
            raise ValueError("Cannot import '{}' from file '{}': file contains wrong {} type '{}'!"
                             .format(cls.__name__, filename, cls._get_object_type_name(),
                                     _object.__class__.__name__
            ))
        return _object

    def to_file(self, filename, file_format=None):
        """Write kafe2 object to file"""
        from ..representation import get_writer

        _format = self._get_file_format(filename=filename, file_format=file_format)
        _writer_class = get_writer(self._get_object_type_name(), _format)
        _writer_class(self, OutputFileHandle(filename=filename)).write()
